import argparse

from .metadata import (
    __version__,
    __summary__,
    __program__
)

from .analyzer import run, MAX_INTERVAL, MIN_INTERVAL
from .dataformat import SUPPORTED_PLATFORMS, SUPPORTED_PLATFORMS_SHORTHANDS

def check_interval(interval):
    """
    Based on the MAX_INTERVAL and MIN_INTERVAL from chat_analyzer.py,
    ensure that the entered interval respects that boundary"""
    interval = int(interval)
    if interval < MIN_INTERVAL:
        raise argparse.ArgumentTypeError(f"Interval must be at least {MIN_INTERVAL} and at most {MAX_INTERVAL}")
    if interval > MAX_INTERVAL:
        raise argparse.ArgumentTypeError(f"Interval must be at most {MAX_INTERVAL} and at least {MIN_INTERVAL}")
    return interval

def check_positive_int(value):
    """
    Check that the value is a positive integer"""
    value = int(value)
    if value < 0:
        raise argparse.ArgumentTypeError("Value must be a positive integer")
    return value

def check_percentile_float(value):
    """
    Check that the value is between 0 and 100 exclusive"""
    value = float(value)
    if value <= 0 or value >= 100:
        raise argparse.ArgumentTypeError("Percentile must be between 0 and 100 exclusive")
    return value


# Anecdotally determined
# For messages that wrap w/ R|, this helps keep spacing consistent. No good way to access indent from argparser
# INDENT_INCREMENT = 2
# Standard help position is 2*INDENT_INCREMENT according to the argparse src
HELP_INDENT_POSITION = 3

# Constants for the argparse get help string
# SUPPRESS = '==SUPPRESS=='
# ZERO_OR_MORE = '*'
# OPTIONAL = '?'
class SmartFormatter(argparse.ArgumentDefaultsHelpFormatter):
    """
    Any help string starting with 'R|' has its newlines (\n) preserved, in addition to
    keeping the fxnality from the HelpFormatter (displaying defaults next to descriptions).
    
    Adapted from and thanks to:
    https://stackoverflow.com/questions/3853722/how-to-insert-newlines-on-argparse-help-text
    """

    # def __init__(self) :
    #     super().__init__(indent_increment=INDENT_INCREMENT)

    def _split_lines(self, text, width):
        if text.startswith('R|'):
            lineList = text[2:].splitlines() 
            # Remove whitespace from before each of the strings so nice spacing doens't affect output inset
            lineList = [line.lstrip() for line in lineList]
            # NOTE: Not using textwrap.wrap here like in argparse because it messes w escape characters. Anecdotally determined indent position instead:
            import textwrap
            lineList = [str(textwrap.fill(line, width, subsequent_indent="\t"*HELP_INDENT_POSITION)) for line in lineList]
            return lineList
        # this is the RawTextHelpFormatter._split_lines
        return argparse.HelpFormatter._split_lines(self, text, width)
    
    # def _get_help_string(self, action):
    #     help = action.help
    #     print("before key error?")
    #     if '%(default)' not in action.help:
    #         print("after key error?")
    #         if action.default is not SUPPRESS:
    #             defaulting_nargs = [OPTIONAL, ZERO_OR_MORE]
    #             if action.option_strings or action.nargs in defaulting_nargs:
    #                 help += ' \033[1m(DEFAULT: %(default)s)\033[0m'
    #     return help

def main():
    parser = argparse.ArgumentParser(description=__summary__, formatter_class=SmartFormatter)

    # Parser info
    parser.prog = __program__
    parser._positionals.title = 'Required Arguments'
    parser._optionals.title = 'Optional Arguments'
    # parser._subparsers.title = 'Subcommands'

    # Meta-arguments
    parser.add_argument('--version', action='version', version=__version__)

    # Required arguments
    parser.add_argument("source", type=str,\
        help=f"""R|
        Raw chat data to process and analyze, or processed sample data to re-analyze.

        In mode=\033[1m'url'\033[0m, (default) source is a url to a past stream/VOD. 
        We currently only support links from: {', '.join(SUPPORTED_PLATFORMS)}. The link must
        be to the original stream/VOD with that attached chat replay. 

        In mode=\033[1m'chatfile'\033[0m, source is a filepath to a .json containing \033[3mraw chat data\033[0m, 
        produced by Xenonva's chat-downloader, or by this program's `--save-chatfile-output` flag. NOTE: the --platform argument is required
        when using this mode.

        In mode=\033[1m'reanalyze'\033[0m, source is a filepath to a .json file previously produced by this program which contains \033[3mexisting sample data to reanalyze\033[0m.
        (Highlights and spikes are regenerated, the existing samples are not affected).""")

    parser.add_argument("--platform", type=str, choices=dict.keys(SUPPORTED_PLATFORMS_SHORTHANDS), help="""
    When reading from a chatfile, specify the platform the chat was downloaded from. 
    By default, Xenova's chat downloader does not store this information with the chat data so it must be manually specified. 
    If the incorrect platform is entered, site-specific data will be innacurate but the program will still run and analyze common attributes.""")

    # Mode arguments
    mode_group = parser.add_argument_group("Program Behavior (Mode)")
    # mutex_mode_group = mode_group.add_mutually_exclusive_group()
    mode_group.add_argument("--mode", "-m", default="url", choices=["url", "chatfile", "reanalyze"], type=str,\
        help="""R|The program can be run in three modes:

        \033[3mNOTE: All modes result in chat analytics output as a .json file.\033[0m

        \033[1m\'url\'\033[0m mode (default) downloads raw chat data from an appropriate source, processes the raw chat data into samples, and then analyzes the samples.

        \033[1m\'chatfile\'\033[0m mode reads raw chat data from a .json file, processes the raw chat data into samples, and then analyzes the samples.
        (We accept raw chat files produced by Xenonva's chat-downloader, or by this program through '--save-chatfile-output').

        \033[1m\'reanalyze\'\033[0m mode reads existing sample data from a .json file produced by this program in a previous run, and recalculates ONLY the post-processed data based on the existing samples. 
        (Highlights and spikes are regenerated, the existing samples are not affected).
        \n""")
    # TODO: Do we restirct use with mode=chatfile, or just admit that it creates a duplicate/slightly different file?
    # TODO: Can't use with reanalyze mode because we don't have access to the chat data, so maybe we just enforce that its a url-only command
    # TODO: Add argument to specify output of the saved chatfile
    mode_group.add_argument("--save-chatfile-output", "-sc", type=str, help="""
    Filepath of the raw chat data to save. If downloading chat data from a URL, 
    save the raw chat data to the provided filepath in addition to processing it, 
    so that the raw data can be \033[3mfully\033[0m reprocessed and analyzed again quickly (using mode='chatfile').
    NOTE: Chatfiles are *much* larger in comparison to the analytics file.  
    NOTE: json file extension is enforced because it affects the content that the chat downloader writes to the file.""")

    # Processing Arguments
    sampling_group = parser.add_argument_group("Processing (Sampling)")
    sampling_group.add_argument("--interval", "-i" , default=5, type=check_interval, help="""
            The time interval (in seconds) at which to compress datapoints into samples. i.e. Duration of the samples. The smaller the interval, the more 
            granular the analytics are. At interval=5, each sample contains 5 seconds of cumulative data.
            *(With the exception of the last sample, which may be shorter than the interval).*""")
    sampling_group.add_argument("--print-interval", default=100, type=int, help="Number of messages between progress updates to the console. If <= 0, progress is not printed.")
    
    
    # Post Processing (Analyzing) Arguments
    postprocess_group = parser.add_argument_group("Post Processing (Analyzing)")
    postprocess_group.add_argument("--highlight-percentile", "-ep", default=93.0, type=check_percentile_float, help="""
    A number between 0 and 100, representing the cutoff percentile that a sample's attribute must meet to be considered a 'highlight' of the chatlog. 
    Samples in the top HIGHLIGHT_PERCENTILE%% of the selected highlight metric will be considered high-engagement samples and included within the constructed highlights. 
    The larger the percentile, the greater the metric requirement before being reported. If 'highlight-percentile'=93.0, only samples in the 93rd percentile (top 7.0%%) of the selected metric will be included in the highlights.
    """)
    metric_choices = ["usersPSec","chatsPSec","activityPSec"] # update metric_to_field map when adding new metrics
    postprocess_group.add_argument("--highlight-metric", "-em", default=metric_choices[0], choices=metric_choices, type=str, help=f"""R|\
    The metric to use for engagement analysis when constructing highlights. Samples in the top HIGHLIGHT_PERCENTILE%% of the selected metric will be considered high-engagement samples and included within the constructed highlights. 
    Each highlight metric choice corresponds to a datapoint for each sample. \n
    \033[1m\'{metric_choices[0]}\'\033[0m compares samples based off of the average number unique users that send a chat per second of the sample.\n
    \033[1m\'{metric_choices[1]}\'\033[0m compares samples based off of the average number of chats per second of the sample (not necessarily sent by unique users).\n 
    \033[1m\'{metric_choices[2]}\'\033[0m compares samples based off of the average number of any type of message that appears in the chat per second of the sample.\n 
    """)

    # TODO: spike sensitivity
    # TODO: spike metric

    # mutex_postprocess_group = postprocess_group.add_mutually_exclusive_group()
     # mutex_postprocess_group.add_argument("--spike-time", default=120, type=check_positive_int, help="Specify the total amount of cumulative spike time (in seconds) that we want output.")
    # TODO: More settings for finding spike. Sensitivity based, or "top-5 based" or...?

    # TODO: Allow user to specify which field they want spike detection on (possibly multiple fields resulting in multiple types of spikes in the list)

   
    # Output Arguments
    output_group = parser.add_argument_group("Output")
    output_group.add_argument("--description", "-d" , type=str, help="R|A description included in the output file to help distinguish it from other output files.\nex: -d \"Ludwig product announcement, small intervals\"")
    output_group.add_argument("--output", "-o", type=str, help="""The filepath to write the output to. If not specified, the output is written to '[MEDIA TITLE].json.' 
                                                    If the provided file path does not end in '.json', the '.json' file extension is appended automaticaly to the filepath (disable with --nojson).""")
    output_group.add_argument("--nojson", action="store_true", help="Disable the automatic appending of the '.json' file extension to the provided output filepath.")
    # TODO: Add a console output group (verbose, quiet, progress update, etc...)

    debug = parser.add_argument_group('Debugging')
    debug.add_argument('--debug','-db', action='store_true', help='Enable debug mode (debug info is printed)')
    debug.add_argument('--break','-b', type=int, default=-1, help='Stop processing messages after BREAK number of messages have been processed. No effect if val < 0')

    args = parser.parse_args()
    kwargs = args.__dict__

    # Argument dependency-checks:
    if(kwargs['mode'] == 'chatfile' and kwargs['platform']== None):
        parser.error('When reading from a chatfile, you must specify the platform the chatfile is from with --platform argument.')
        # TODO: Add description of this to the chatfile desc and add the actual platform arg itself
    if(kwargs['save_chatfile_output']):  
        if(kwargs['mode'] != 'url'):
            parser.error('The --save-chatfile-output flag can only be used in mode=\033[1m\'url\'\033[0m.')
        if(not kwargs['save_chatfile_output'].endswith('.json')):
            kwargs['save_chatfile_output'] += '.json'
    if(kwargs['output']):
        if(not kwargs['output'].endswith('.json') and not kwargs['nojson']):
            kwargs['output'] += '.json'
    # TODO: Interval should not be allowed with mode = reanalyze (clarify again what reanalyze is for)
    # TODO: Low percentiles result in no highlights... why?

    run(**kwargs)




# Some testing URLs
# url = 'https://www.youtube.com/watch?v=97w16cYskVI' # yt stream that comes with lots of message types (retrieved from chat-downloader testing sample) TODO: [blocked now?! check into]
# url = 'asdds.com/a/b/c/d' # (error) invalid URL
# url = 'https://www.youtube.com/watch?v=5qap5aO4i9A' # (error) stream still live (lo-fi hip hop girl runs 24/7)
# url = 'https://www.twitch.tv/videos/1522574868'  # summit1g's 14 hour stream
# url = 'https://www.youtube.com/watch?v=PTWpoZITraE&ab_channel=RobScallon' # (error) Youtube video without chat replay
# url = 'https://www.youtube.com/watch?v=UR902_1LhVk&t=24333s&ab_channel=Ludwig' # Ludwig's 1 million dollar game poker stream, 8:57:25, 158366 totalActivity
# # url = 'https://www.youtube.com/watch?v=vjBNozL9Daw' #(error for now TODO: test later) no chat replay
# url = 'https://www.twitch.tv/videos/1289325547' # markiplier peen stream
# url = 'https://www.twitch.tv/videos/1530042943' # MMG's stream
# url = 'https://clips.twitch.tv/AverageSparklyTortoisePeoplesChamp' # (error) chat replay not avail
# url = 'https://www.twitch.tv/videos/1534993737' # Huge fkin XQC stream

# TODO: Check the one below for type: sponsorships_gift_redemption_announcement
# TODO: Look into sponsorships_gift_redemption_announcement (...'was gifted a membership by...')
# url = 'https://www.youtube.com/watch?v=1jRVuFcBj3M&list=PLLGT0cEMIAzcd5XagsMwz22-NFPmToKIP&index=2&ab_channel=Ludwig'

# TODO: Consider 'raid' type:
# url = 'https://www.twitch.tv/videos/1538666427'


# chat_analyzer 'https://www.twitch.tv/videos/1522574868' --print-interval 100 -i 10


# chat_analyzer 'https://www.twitch.tv/videos/1289325547' -o output/analyze/markiplier_peen.json -sc output/raw/markiplier_peen_chat.json
# chat_analyzer output/analyze/markiplier_peen.json -m reanalyze -o output/analyze/markiplier_peen_reanalyzed.json


"""Set metadata for chat-downloader"""

__title__ = 'chat-downloader'
__program__ = 'chat_downloader'
__summary__ = 'A simple tool used to retrieve chat messages from livestreams, videos, clips and past broadcasts. No authentication needed!'
__author__ = 'xenova'
__email__ = 'admin@xenova.com'
__copyright__ = '2020-2022 xenova'
__url__ = 'https://github.com/xenova/chat-downloader'
__version__ = '0.2.0'



"""Console script for chat_downloader."""
import argparse
import re
from docstring_parser import parse as doc_parse


from .chat_downloader import (
    ChatDownloader,
    run
)

from .metadata import (
    __version__,
    __summary__,
    __program__
)

from .utils.core import (
    get_default_args,
    int_or_none
)

from .debugging import (
    disable_logger,
    set_log_level
)


def splitter(s):
    return [item.strip() for item in re.split(r'[\s,;]+', s)]


def str2bool(value):
    if isinstance(value, bool):
        return value
    value = value.lower()
    if value in ('true', 'yes',  't', 'y', '1', 'enable'):
        return True
    elif value in ('false', 'no', 'f', 'n', '0', 'disable'):
        return False
    else:
        raise argparse.ArgumentTypeError(
            f'Boolean value expected: {value} is not a boolean')


def main(cli_args=None):

    parser = argparse.ArgumentParser(
        description=__summary__,
        # formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.prog = __program__

    parser.add_argument('--version', action='version', version=__version__)

    def get_info(function):
        info = {}

        docstring = doc_parse(function.__doc__)
        args = get_default_args(function)

        for param in docstring.params:
            info[param.arg_name] = {
                'help': param.description,
                'default': args.get(param.arg_name)
            }
        return info

    # get help and default info
    get_chat_info = get_info(ChatDownloader.get_chat)
    get_init_info = get_info(ChatDownloader.__init__)

    def add_param(param_type, group, *keys, **kwargs):
        info = get_chat_info if param_type == 'chat' else get_init_info
        key = keys[0].lstrip('-')
        group.add_argument(*keys, **info[key], **kwargs)

    def add_chat_param(group, *keys, **kwargs):
        add_param('chat', group, *keys, **kwargs)

    def add_init_param(group, *keys, **kwargs):
        add_param('init', group, *keys, **kwargs)

    add_chat_param(parser, 'url')

    time_group = parser.add_argument_group('Timing Arguments')

    add_chat_param(time_group, '--start_time', '-s')
    add_chat_param(time_group, '--end_time', '-e')

    # Specify message types/groups
    type_group = parser.add_argument_group('Message Type Arguments')
    type_options = type_group.add_mutually_exclusive_group()

    add_chat_param(type_options, '--message_types', type=splitter)
    add_chat_param(type_options, '--message_groups', type=splitter)

    retry_group = parser.add_argument_group(
        'Retry Arguments')  # what to do when an error occurs
    add_chat_param(retry_group, '--max_attempts', type=int)
    add_chat_param(retry_group, '--retry_timeout', type=float)
    add_chat_param(retry_group, '--interruptible_retry',
                   type=str2bool, nargs='?', const=True)

    termination_group = parser.add_argument_group('Termination Arguments')
    add_chat_param(termination_group, '--max_messages', type=int)
    add_chat_param(
        termination_group, '--inactivity_timeout', type=float)
    add_chat_param(termination_group, '--timeout', type=float)

    # TODO request_timeout
    # specify how long to spend on any single http request

    # TODO
    # parser.add_argument('--force_encoding', default=default_params['force_encoding'],
    #                     help='force certain encoding\n(default: %(default)s)')

    # Formatting
    format_group = parser.add_argument_group('Format Arguments')
    # format_group.add_argument('--print_json', action='store_true', help='Print out json ', default=None)
    add_chat_param(format_group, '--format')
    add_chat_param(format_group, '--format_file')

    # info = get_site_info(YouTubeChatDownloader)
    youtube_group = parser.add_argument_group(
        '[Site Specific] YouTube Arguments')
    add_chat_param(youtube_group, '--chat_type',
                   choices=['live', 'top'])
    add_chat_param(youtube_group, '--ignore', type=splitter)
    # add_chat_param(
    #     youtube_group, '--force_no_timeout', action='store_true')

    # info = get_site_info(TwitchChatDownloader)
    twitch_group = parser.add_argument_group(
        '[Site Specific] Twitch Arguments')
    add_chat_param(
        twitch_group, '--message_receive_timeout', type=float)
    add_chat_param(twitch_group, '--buffer_size', type=int)

    output_group = parser.add_argument_group('Output Arguments')
    add_chat_param(output_group, '--output', '-o')
    add_chat_param(output_group, '--overwrite',
                   type=str2bool, nargs='?', const=True)
    add_chat_param(output_group, '--sort_keys',
                   type=str2bool, nargs='?', const=True)
    add_chat_param(output_group, '--indent', type=lambda x: int_or_none(x, x))

    # Debugging only available from the CLI
    debug_group = parser.add_argument_group('Debugging/Testing Arguments')

    on_debug_options = debug_group.add_mutually_exclusive_group()
    on_debug_options.add_argument('--pause_on_debug', action='store_true',
                                  help='Pause on certain debug messages, defaults to False')
    on_debug_options.add_argument('--exit_on_debug', action='store_true',
                                  help='Exit when something unexpected happens, defaults to False')

    debug_options = debug_group.add_mutually_exclusive_group()
    debug_options.add_argument('--logging', choices=['none', 'debug', 'info', 'warning', 'error', 'critical'],
                               help='Level of logging to display, defaults to info', default='info')

    debug_options.add_argument('--testing', action='store_true',
                               help='Enable testing mode. This is equivalent to setting logging to debug and enabling pause_on_debug. Defaults to False')
    debug_options.add_argument('--verbose', '-v', action='store_true',
                               help='Print various debugging information. This is equivalent to setting logging to debug. Defaults to False')
    debug_options.add_argument('--quiet', '-q', action='store_true',
                               help='Activate quiet mode (hide all output), defaults to False')

    # INIT PARAMS
    init_group = parser.add_argument_group('Initialisation Arguments')
    add_init_param(init_group, '--cookies', '-c')
    add_init_param(init_group, '--proxy', '-p')

    # TODO add headers (user agent) as arg

    parser._positionals.title = 'Mandatory Arguments'
    parser._optionals.title = 'General Arguments'

    args = parser.parse_args(args=cli_args)

    # Modify debugging args:
    if args.testing:  # (only for CLI)
        args.logging = 'debug'
        args.pause_on_debug = True

    if args.verbose:
        args.logging = 'debug'

    if args.quiet or args.logging == 'none':
        disable_logger()
    else:
        set_log_level(args.logging)

    # Run with these arguments
    run(**args.__dict__)



"""Main module."""
import sys
import itertools
import time
import json

from urllib.parse import urlparse

from .metadata import __version__

from .sites.common import (
    SiteDefault,
    BaseChatDownloader
)
from .sites import get_all_sites

from .formatting.format import ItemFormatter
from .utils.core import (
    safe_print,
    get_default_args,
    update_dict_without_overwrite
)

from .utils.timed_utils import TimedGenerator

from .debugging import (
    log,
    set_testing_mode,
    TestingModes,
    TestingException
)

from .output.continuous_write import ContinuousWriter


from requests.exceptions import (
    RequestException,
    ConnectionError
)

from .errors import (
    URLNotProvided,
    SiteNotSupported,
    InvalidURL,
    ChatDownloaderError,
    ChatGeneratorError,
    ParsingError
)


class ChatDownloader():
    """Class used to create sessions and download chats."""

    def __init__(self,
                 headers=None,
                 cookies=None,
                 proxy=None,
                 ):
        """Initialise a new session for making requests. Parameters are saved
        and are sent to the relevant constructor when creating a new session.

        :param headers: Headers to use for subsequent requests, defaults to None
        :type headers: dict, optional
        :param cookies: Path of cookies file, defaults to None
        :type cookies: str, optional
        :param proxy: Use the specified HTTP/HTTPS/SOCKS proxy. To enable SOCKS
            proxy, specify a proper scheme. For example socks5://127.0.0.1:1080/.
            Pass in an empty string (--proxy "") for direct connection. Defaults
            to None
        :type proxy: str, optional
        """

        self.init_params = locals()
        self.init_params.pop('self')

        log('debug', f'Python version: {sys.version}')
        log('debug', f'Program version: {__version__}')
        log('debug', f'Initialisation parameters: {self.init_params}')

        # Track sessions using a dictionary (allows for reusing)
        self.sessions = {}

    def get_chat(self, url=None,
                 start_time=None,
                 end_time=None,
                 max_attempts=15,
                 retry_timeout=None,
                 interruptible_retry=True,
                 timeout=None,
                 inactivity_timeout=None,
                 max_messages=None,

                 message_groups=SiteDefault('message_groups'),
                 message_types=None,

                 # Output
                 output=None,
                 overwrite=True,
                 sort_keys=True,
                 indent=4,

                 # Formatting
                 format=SiteDefault('format'),
                 format_file=None,

                 # YouTube
                 chat_type='live',
                 ignore=None,

                 # Twitch
                 message_receive_timeout=0.1,
                 buffer_size=4096
                 ):
        """Used to get chat messages from a livestream, video, clip or past broadcast.

        :param url: The URL of the livestream, video, clip or past broadcast,
            defaults to None
        :type url: str, optional
        :param start_time: Start time in seconds or hh:mm:ss, defaults
            to None (as early as possible)
        :type start_time: float, optional
        :param end_time: End time in seconds or hh:mm:ss, defaults to
            None (until the end)
        :type end_time: float, optional
        :param max_attempts: Maximum number of attempts to retrieve chat
            messages, defaults to 15
        :type max_attempts: int, optional
        :param retry_timeout: Number of seconds to wait before retrying. Setting
            this to a negative number will wait for user input. Default is None
            (use exponential backoff, i.e. immediate, 1s, 2s, 4s, 8s, ...)
        :type retry_timeout: float, optional
        :param interruptible_retry: Have the option to skip waiting and
            immediately retry. Defaults to True
        :type interruptible_retry: bool, optional
        :param timeout: Stop retrieving chat after a certain duration
            (in seconds), defaults to None
        :type timeout: float, optional
        :param inactivity_timeout: Stop getting messages after not receiving
            anything for a certain duration (in seconds), defaults to None
        :type inactivity_timeout: float, optional
        :param max_messages: Maximum number of messages to retrieve, defaults
            to None (unlimited)
        :type max_messages: int, optional
        :param message_groups: List of messages groups (a predefined,
            site-specific collection of message types) to include
        :type message_groups: SiteDefault, optional
        :param message_types: List of messages types to include, defaults to None
        :type message_types: list, optional
        :param output: Path of the output file, defaults to None (print to
            standard output)
        :type output: str, optional
        :param overwrite: If True, overwrite output file. Otherwise, append
            to the end of the file. Defaults to True. In both cases, the file
            (and directories) is created if it does not exist.
        :type overwrite: bool, optional
        :param sort_keys: Sort keys when outputting to a file, defaults to True
        :type sort_keys: bool, optional
        :param indent: Number of spaces to indent JSON objects by. If
            nonnumerical input is provided, this will be used to indent
            the objects. Defaults to 4
        :type indent: Union[int, str], optional
        :param format: Specify how messages should be formatted for printing,
            defaults to the site's default value
        :type format: SiteDefault, optional
        :param format_file: Specify the path of the format file to choose
            formats from, defaults to None
        :type format_file: str, optional
        :param chat_type: Specify chat type, defaults to 'live'
        :type chat_type: str, optional
        :param ignore: Ignore a list of video ids, defaults to None
        :type ignore: list, optional
        :param message_receive_timeout: Time before requesting for new messages,
            defaults to 0.1
        :type message_receive_timeout: float, optional
        :param buffer_size: Specify a buffer size for retrieving messages,
            defaults to 4096
        :type buffer_size: int, optional
        :raises URLNotProvided: if no URL is provided
        :raises ChatGeneratorError: if no valid generator can be found for a site
        :raises SiteNotSupported: if no matching site can be found
        :raises InvalidURL: if the URL provided is not valid
        :return: The appropriate Chat object, given these parameters
        :rtype: Chat
        """

        # TODO params to add
        # If True, program will not sleep when a timeout instruction is given
        # force_no_timeout=False,
        # :param force_no_timeout: Force no timeout between subsequent requests
        #  force_encoding=None, # use default

        if not url:
            raise URLNotProvided('No URL provided.')

        original_params = locals()
        original_params.pop('self')

        # loop through all websites and
        # get corresponding website parser
        # based on matching url with predefined regex
        for site in get_all_sites():
            match_info = site.matches(url)
            if match_info:  # match found

                function_name, match = match_info

                # Create new session
                self.create_session(site)
                site_object = self.sessions[site.__name__]

                # Parse site-defaults
                params = {}
                for k, v in original_params.items():
                    params[k] = site_object.get_site_value(v)

                log('info', f'Site: {site_object._NAME}')
                log('debug', f'Program parameters: {params}')

                get_chat = getattr(site_object, function_name, None)
                if not get_chat:
                    raise NotImplementedError(
                        f'{function_name} has not been implemented in {site.__name__}.')

                chat = get_chat(match, params)
                log('debug',
                    f'Match found: "{match}". Running "{function_name}" function in "{site.__name__}".')

                if chat is None:
                    raise ChatGeneratorError(
                        f'No valid generator found in {site.__name__} for url "{url}"')

                if isinstance(params['max_messages'], int):
                    chat.chat = itertools.islice(
                        chat.chat, params['max_messages'])
                else:
                    pass  # TODO throw error

                if params['timeout'] is not None or params['inactivity_timeout'] is not None:
                    # Generator requires timing functionality

                    chat.chat = TimedGenerator(
                        chat.chat, params['timeout'], params['inactivity_timeout'])

                    if isinstance(params['timeout'], (float, int)):
                        start = time.time()

                        def log_on_timeout():
                            log('debug',
                                f'Timeout occurred after {time.time() - start} seconds.')
                        setattr(chat.chat, 'on_timeout', log_on_timeout)

                    if isinstance(params['inactivity_timeout'], (float, int)):
                        def log_on_inactivity_timeout():
                            log('debug',
                                f"Inactivity timeout occurred after {params['inactivity_timeout']} seconds.")
                        setattr(chat.chat, 'on_inactivity_timeout',
                                log_on_inactivity_timeout)

                formatter = ItemFormatter(params['format_file'])
                chat.format = lambda x: formatter.format(
                    x, format_name=params['format'])

                if params['output']:
                    chat.attach_writer(ContinuousWriter(
                        params['output'],
                        indent=params['indent'],
                        sort_keys=params['sort_keys'],
                        overwrite=params['overwrite'],
                        lazy_initialise=True
                    ))

                chat.site = site_object

                log('debug', f'Chat information: {chat.__dict__}')
                log('info', f'Retrieving chat for "{chat.title}".')

                return chat

        parsed = urlparse(url)
        log('debug', str(parsed))

        if parsed.netloc:
            raise SiteNotSupported(f'Site not supported: {parsed.netloc}')
        elif not parsed.scheme:  # No scheme, try to correct
            original_params['url'] = 'https://' + url
            chat = self.get_chat(**original_params)
            if chat:
                return chat
        else:
            raise InvalidURL(f'Invalid URL: "{url}"')

    def create_session(self, chat_downloader_class, overwrite=False):
        if not issubclass(chat_downloader_class, BaseChatDownloader):
            raise TypeError(
                f'Unable to create session, class must extend BaseChatDownloader. Class given: {chat_downloader_class}')
        elif chat_downloader_class == BaseChatDownloader:
            raise TypeError(
                'Unable to create session, class may not be BaseChatDownloader.')

        session_name = chat_downloader_class.__name__
        log('debug', f'Created {session_name} session.')

        if session_name not in self.sessions or overwrite:
            self.sessions[session_name] = chat_downloader_class(
                **self.init_params)

        return self.sessions[session_name]

    def get_session(self, chat_downloader_class):
        return self.sessions.get(chat_downloader_class.__name__)

    def close(self):
        """Close all sessions associated with the object"""
        for session in self.sessions.values():
            session.close()

        self.sessions = {}


def run(propagate_interrupt=False, **kwargs):
    """
    Create a single session and get the chat using the specified parameters.
    """

    # Set testing mode
    if kwargs.get('exit_on_debug'):
        set_testing_mode(TestingModes.EXIT_ON_DEBUG)
    elif kwargs.get('pause_on_debug'):
        set_testing_mode(TestingModes.PAUSE_ON_DEBUG)

    init_param_names = get_default_args(ChatDownloader.__init__)
    program_param_names = get_default_args(ChatDownloader.get_chat)

    update_dict_without_overwrite(kwargs, init_param_names)
    update_dict_without_overwrite(kwargs, program_param_names)

    chat_params = {}
    init_params = {}

    for arg in kwargs:
        value = kwargs[arg]

        print(f"analyzing arg: {value}")

        if arg in program_param_names:
            chat_params[arg] = value
            print(f"{value} added to chat_params")
        elif arg in init_param_names:
            init_params[arg] = value

    downloader = ChatDownloader(**init_params)

    try:
        chat = downloader.get_chat(**chat_params)

        if kwargs.get('quiet'):  # Only check if quiet once
            def callback(item):
                pass
        else:
            def callback(item):
                chat.print_formatted(item)

        for message in chat:
            callback(message)

        log('info', 'Finished retrieving chat messages.')

    except (
        ChatGeneratorError,
        ParsingError,
        TestingException
    ) as e:  # Errors which may be bugs
        log('error', f'{e}. Please report this at https://github.com/xenova/chat-downloader/issues/new/choose')

    except ChatDownloaderError as e:  # Expected errors
        log('error', e)

    except ConnectionError as e:
        log(
            'error', f'Unable to establish a connection. Please check your internet connection. {e}')

    except RequestException as e:
        log('error', e)

    except KeyboardInterrupt as e:
        if propagate_interrupt:
            raise e
        else:
            log('error', 'Keyboard Interrupt')

    finally:
        downloader.close()



from .common import (
    Chat,
    BaseChatDownloader,
    Remapper as r,
    Image
)

from ..errors import (
    SiteError,
    NoChatReplay,
    VideoUnavailable,
    UserNotFound
)

from ..utils.core import (
    ensure_seconds,
    timestamp_to_microseconds,
    seconds_to_time,
    int_or_none,
    str_or_none,
    replace_with_underscores,
    multi_get,
    remove_prefixes,
    attempts
)

from ..debugging import (
    log,
    debug_log
)

import re
import json
import time
import socket
import base64
import math
from requests.exceptions import RequestException
from json.decoder import JSONDecodeError


class TwitchError(SiteError):
    """Raised when an error occurs with a Twitch video."""
    pass


# TODO export as another module?


class TwitchChatIRC():

    def __init__(self):
        # create new socket
        self.socket = socket.socket()

        # start connection
        self.socket.connect(('irc.chat.twitch.tv', 6667))
        # print('Connected to', self._HOST, 'on port', self._PORT)

        self.current_channel = None
        # https://dev.twitch.tv/docs/irc/tags
        # https://dev.twitch.tv/docs/irc/membership
        # https://dev.twitch.tv/docs/irc/commands

        # twitch.tv/membership
        self.send_raw(
            'CAP REQ :twitch.tv/tags twitch.tv/commands twitch.tv/membership')
        self.send_raw('PASS SCHMOOPIIE')
        self.send_raw('NICK justinfan67420')

    def send_raw(self, string):
        self.socket.send((string + '\r\n').encode('utf-8'))

    def recv(self, buffer_size):
        return self.socket.recv(buffer_size).decode('utf-8', 'ignore')

    def join_channel(self, channel_name):
        channel_lower = channel_name.lower()

        if self.current_channel != channel_lower:
            self.send_raw(f'JOIN #{channel_lower}')
            self.current_channel = channel_lower

    def set_timeout(self, message_receive_timeout):
        self.socket.settimeout(message_receive_timeout)

    def close_connection(self):
        self.socket.close()


class TwitchChatDownloader(BaseChatDownloader):
    _BADGE_INFO = {}
    _BADGE_INFO_URL = 'https://badges.twitch.tv/v1/badges/global/display'
    # TODO add local version of badge list?

    _NAME = 'twitch.tv'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Only get badge info if not already retrieved
        # TODO add argument (no_badges)
        if not TwitchChatDownloader._BADGE_INFO:
            TwitchChatDownloader._BADGE_INFO = self._session_get_json(
                self._BADGE_INFO_URL).get('badge_sets') or {}

    _TESTS = [
        # Live
        {
            'name': 'Livestream',
            'params': {
                'url': 'https://www.twitch.tv/xenova',
                'timeout': 5
            }
        },

        # Past broadcasts
        {
            'name': 'Past broadcast with chat replay.',
            'params': {
                'url': 'https://www.twitch.tv/videos/87136772',
                'max_messages': 30
            },
            'expected_result': {
                'message_types': ['text_message'],
                # 'action_types': [],
                'messages_condition': lambda messages: len(messages) <= 30
            }
        },

        # Clip
        {
            'name': 'Clip with chat replay.',
            'params': {
                'url': 'https://clips.twitch.tv/TrappedFrigidPenguinSeemsGood',
            },
            'expected_result': {
                'message_types': ['text_message'],
                # 'action_types': [],
                'messages_condition': lambda messages: len(messages) > 0
                # 'error': LoginRequired,
            }
        },


        # Commenter is "None"
        # https://www.twitch.tv/videos/873616984 --start_time 1:26:10


        {
            'name': "This clip's past broadcast has expired and chat replay is no longer available.",
            'params': {
                'url': 'https://clips.twitch.tv/AverageSparklyTortoisePeoplesChamp',
            },
            'expected_result': {
                'error': NoChatReplay
            }
        },
        {
            'name': "Sorry. Unless you've got a time machine, that content is unavailable.",
            'params': {
                'url': 'https://www.twitch.tv/videos/1',
            },
            'expected_result': {
                'error': VideoUnavailable
            }
        },
    ]

    # clips
    # vod
    # name -> live

    _SITE_DEFAULT_PARAMS = {
        'format': 'twitch',
    }

    _VALID_URLS = {
        # e.g. 'http://www.twitch.tv/riotgames/v/6528877?t=5m10s'
        '_get_chat_by_vod_id': r'''(?x)
                    https?://
                        (?:
                            (?:(?:www|go|m)\.)?twitch\.tv/(?:[^/]+/v(?:ideo)?|videos)/|
                            player\.twitch\.tv/\?.*?\bvideo=v?
                        )
                        (?P<id>\d+)
                    ''',

        # e.g. 'https://clips.twitch.tv/FaintLightGullWholeWheat'
        '_get_chat_by_clip_id': r'''(?x)
                        https?://
                            (?:
                                clips\.twitch\.tv/(?:embed\?.*?\bclip=|(?:[^/]+/)*)|
                                (?:(?:www|go|m)\.)?twitch\.tv/[^/]+/clip/
                            )
                            (?P<id>[^/?#&]+)
                        ''',

        # e.g. 'http://www.twitch.tv/shroomztv'
        '_get_chat_by_stream_id': r'''(?x)
                        https?://
                            (?:
                                (?:(?:www|go|m)\.)?twitch\.tv/|
                                player\.twitch\.tv/\?.*?\bchannel=
                            )
                            (?P<id>[^/#?]+)
                        '''
    }

    _CLIENT_ID = 'kimne78kx3ncx6brgo4mv6wki5h1ko'  # public client id

    _GQL_API_URL = 'https://gql.twitch.tv/gql'
    _API_TEMPLATE = 'https://api.twitch.tv/v5/videos/{}/comments?client_id={}'

    _PING_TEXT = 'PING :tmi.twitch.tv'
    _PONG_TEXT = 'PONG :tmi.twitch.tv'

    _SUBSCRIPTION_TYPES = {
        'Prime': 'Prime',
        '1000': 'Tier 1',
        '2000': 'Tier 2',
        '3000': 'Tier 3'
    }

    @staticmethod
    def _parse_bool(text):
        return text == '1'

    @staticmethod
    def _parse_bool_text(text):
        return text == 'true'

    @staticmethod
    def _parse_author_images(original_url):
        # e.g. https://static-cdn.jtvnw.net/jtv_user_pictures/3892c956-0616-4fc9-b2fe-527b1be0b623-profile_image-300x300.png
        smaller_icon = original_url.replace('300x300', '70x70')
        return [
            Image(original_url, 300, 300).json(),
            Image(smaller_icon, 70, 70).json(),
        ]

    @staticmethod
    def _parse_commenter(commenter):
        info = {}
        for key in commenter or []:
            r.remap(info, TwitchChatDownloader._AUTHOR_REMAPPING,
                    key, commenter[key])
        return info

    @staticmethod
    def _parse_message_info(message):
        # Ignore: fragments, is_action
        message_text = message.get('body') or ''

        message_emotes = {}
        locations = {}

        for emoticon in message.get('emoticons') or []:
            emote_id = emoticon.get('_id')
            begin = emoticon.get('begin')
            end = emoticon.get('end')

            if emote_id:
                if emote_id not in message_emotes:
                    message_emotes[emote_id] = {
                        'id': emote_id,
                        'images': TwitchChatDownloader._generate_emote_image_list(emote_id),
                        'name': message_text[begin:end + 1]
                    }

                    locations[emote_id] = []

                locations[emote_id].append(f'{begin}-{end}')

        for emote_id in message_emotes:
            message_emotes[emote_id]['locations'] = ','.join(
                locations[emote_id])

        message_info = {
            'message': message_text,
            'author_colour': message.get('user_color'),
            'author_badges': message.get('user_badges') or [],
            'user_notice_params': message.get('user_notice_params') or {},
        }

        if message_emotes:
            message_info['emotes'] = list(message_emotes.values())
            # TwitchChatDownloader._add_text_for_emotes(message_info['message'], message_info['emotes'])

        return message_info

    @staticmethod
    def _decode_pseudo_BNF(text):
        """
        Decode text according to https://ircv3.net/specs/extensions/message-tags.html
        """
        return text.replace(r'\:', ';').replace(r'\s', ' ')

    @staticmethod
    def _generate_emote_image_list(emote_id):
        emote_image_list = []
        for theme in ('light', 'dark'):
            for size in ((28, '1.0'), (56, '2.0'), (112, '3.0')):
                image = Image(
                    TwitchChatDownloader._EMOTE_URL_TEMPLATE.format(
                        emote_id, theme, size[1]),
                    size[0],
                    size[0],
                    f'{size[0]}x{size[0]}-{theme}'
                ).json()

                emote_image_list.append(image)
        return emote_image_list

    _EMOTE_REGEX = r'(\d+):([\d,-]+)'
    _EMOTE_URL_TEMPLATE = 'https://static-cdn.jtvnw.net/emoticons/v2/{}/default/{}/{}'

    @staticmethod
    def _parse_emotes(text):
        # Information to replace text in the message with emote images. This can be empty.
        # <emote ID>:<first index>-<last index>,<another first index>-<another last index>/<another emote ID>:<first index>-<last index>
        emotes = []

        matches = re.findall(TwitchChatDownloader._EMOTE_REGEX, text)

        for match in matches:
            emote_id = match[0]
            emote = {
                'id': emote_id,
                'locations': match[1].split(','),
                'images': TwitchChatDownloader._generate_emote_image_list(emote_id)
            }
            emotes.append(emote)

        return emotes

    _AUTHOR_REMAPPING = {
        '_id': r('id', str_or_none),
        'name': 'name',
        'display_name': 'display_name',
        'logo': r('images', _parse_author_images),
        'type': 'type',
        'created_at': r('created_at', timestamp_to_microseconds),
        # 'updated_at': r('updated_at', 'parse_timestamp'),
        'bio': 'bio'
    }

    _COMMENT_REMAPPING = {
        '_id': 'message_id',
        'created_at': r('timestamp', timestamp_to_microseconds),
        'commenter': r('author', _parse_commenter),

        'content_offset_seconds': 'time_in_seconds',

        'source': 'source',
        'state': 'state',
        # TODO make sure body vs. fragments okay
        'message': r(None, _parse_message_info, True)
    }

    _MESSAGE_PARAM_REMAPPING = {
        'msg-id': 'message_type',

        'msg-param-cumulative-months': r('cumulative_months', int_or_none),
        'msg-param-months': r('months', int_or_none),
        'msg-param-displayName': 'raider_display_name',
        'msg-param-login': 'raider_name',
        'msg-param-viewerCount': r('number_of_raiders', int_or_none),

        'msg-param-promo-name': 'promotion_name',
        'msg-param-promo-gift-total': 'number_of_gifts_given_during_promo',

        'msg-param-recipient-id': 'gift_recipient_id',
        'msg-param-recipient-user-name': 'gift_recipient_display_name',
        'msg-param-recipient-display-name': 'gift_recipient_display_name',
        'msg-param-gift-months': r('number_of_months_gifted', int_or_none),


        'msg-param-sender-login': 'gifter_name',
        'msg-param-sender-name': 'gifter_display_name',

        'msg-param-should-share-streak': r('user_wants_to_share_streaks', _parse_bool),
        'msg-param-streak-months': r('number_of_consecutive_months_subscribed', int_or_none),
        'msg-param-sub-plan': r('subscription_type', lambda x: TwitchChatDownloader._SUBSCRIPTION_TYPES.get(x)),
        'msg-param-sub-plan-name': r('subscription_plan_name', _decode_pseudo_BNF),
        'msg-param-sub-benefit-end-month': r('sub_benefit_end_month', int_or_none),

        'msg-param-ritual-name': 'ritual_name',

        'msg-param-threshold': 'bits_badge_tier',


        # found in vods

        # resub
        'msg-param-multimonth-duration': r('multimonth_duration', int_or_none),
        'msg-param-multimonth-tenure': r('multimonth_tenure', int_or_none),
        'msg-param-was-gifted': r('was_gifted', _parse_bool_text),

        'msg-param-gifter-id': 'gifter_id',
        'msg-param-gifter-login': 'gifter_name',
        'msg-param-gifter-name': 'gifter_display_name',
        'msg-param-anon-gift': r('was_anonymous_gift', _parse_bool_text),
        'msg-param-gift-month-being-redeemed': r('gift_months_being_redeemed', int_or_none),

        # rewardgift
        'msg-param-domain': 'domain',
        'msg-param-selected-count': r('selected_count', int_or_none),
        'msg-param-trigger-type': 'trigger_type',
        'msg-param-total-reward-count': r('total_reward_count', int_or_none),
        'msg-param-trigger-amount': r('trigger_amount', int_or_none),

        # submysterygift
        'msg-param-origin-id': r('origin_id', _decode_pseudo_BNF),
        'msg-param-sender-count': r('sender_count', int_or_none),
        'msg-param-mass-gift-count': r('mass_gift_count', int_or_none),

        # communitypayforward
        'msg-param-prior-gifter-anonymous': r('prior_gifter_anonymous', _parse_bool_text),
        'msg-param-prior-gifter-user-name': 'prior_gifter_name',
        'msg-param-prior-gifter-display-name': 'prior_gifter_display_name',
        'msg-param-prior-gifter-id': 'prior_gifter_id',

        'msg-param-fun-string': 'fun_string',

        # charity
        'msg-param-charity': 'charity',
        'msg-param-charity-name': 'charity_name',
        'msg-param-charity-hashtag': 'charity_hashtag',
        'msg-param-charity-learn-more': 'charity_link',
        'msg-param-charity-hours-remaining': r('charity_hours_remaining', int_or_none),
        'msg-param-charity-days-remaining': r('charity_days_remaining', int_or_none),
        'msg-param-total': r('charity_total_raised', int_or_none),

        # not come across yet, but other tools have it:
        # 'msg-param-bits-amount':'bits_amount',
        # 'msg-param-streak-tenure-months':'streak_tenureBaseChatDownloaderBase#
        # 'msg-param-userID':'user_id',
        #
        # 'msg-param-cumulative-tenure-months':'cumulative_tenure_months',
        # 'msg-param-should-share-streak-tenure':'should-share-streak-tenure',
        # 'msg-param-min-cheer-amount' :'minimum_cheer_amount',
        # 'msg-param-gift-name':'gift_name',




        # to remove later
        'msg-param-profileImageURL': 'profile_image_url'
    }

    # Create set of known types
    _KNOWN_COMMENT_KEYS = {
        # created elsewhere
        'message', 'time_in_seconds', 'message_id', 'time_text', 'author', 'timestamp', 'message_type', 'emotes'
    }

    _KNOWN_COMMENT_KEYS.update(BaseChatDownloader.get_mapped_keys({
        **_COMMENT_REMAPPING, **_MESSAGE_PARAM_REMAPPING
    }))
    # print('_KNOWN_COMMENT_KEYS',_KNOWN_COMMENT_KEYS)

    _IRC_REMAPPING = {
        # CLEARCHAT
        # Purges all chat messages in a channel, or purges chat messages from a specific user, typically after a timeout or ban.
        # (Optional) Duration of the timeout, in seconds. If omitted, the ban is permanent.
        'ban-duration': r('ban_duration', int_or_none),

        # CLEARMSG
        # Removes a single message from a channel. This is triggered by the/delete <target-msg-id> command on IRC.
        # Name of the user who sent the message.
        'login': 'author_name',
        # UUID of the message.
        'target-msg-id': 'target_message_id',


        # GLOBALUSERSTATE
        # On successful login, provides data about the current logged-in user through IRC tags. It is sent after successfully authenticating (sending a PASS/NICK command).

        'emote-sets': 'emote_sets',  # TODO split by,?

        # GENERAL
        # can be empty (which means it depends on dark/light theme)
        'color': 'colour',
        'display-name': 'author_display_name',
        'user-id': r('author_id', str_or_none),



        # reply-parent-display-name


        # PRIVMSG
        'badge-info': 'author_badge_metadata',
        'badges': 'author_badges',

        'bits': r('bits', int_or_none),

        'id': 'message_id',
        'mod': r('author_is_moderator', _parse_bool),
        'room-id': r('channel_id', str_or_none),
        'tmi-sent-ts': r('timestamp', lambda x: int_or_none(x, 0) * 1000),

        'subscriber': r('author_is_subscriber', _parse_bool),
        'turbo': r('author_is_turbo', _parse_bool),

        'client-nonce': 'client_nonce',

        'user-type': 'user_type',



        'reply-parent-msg-body': r('in_reply_to_message', _decode_pseudo_BNF),
        'reply-parent-user-id': r('in_reply_to_author_id', str_or_none),

        'reply-parent-msg-id': 'in_reply_to_message_id',
        'reply-parent-display-name': 'in_reply_to_author_display_name',
        'reply-parent-user-login': 'in_reply_to_author_name',

        'crowd-chant-parent-msg-id': 'crowd_chant_in_reply_to_message_id',

        'custom-reward-id': 'custom_reward_id',


        'emotes': r('emotes', _parse_emotes),
        'flags': 'flags',
        'first-msg': r('is_first_message', _parse_bool),



        # ROOMSTATE
        'emote-only': r('emote_only', _parse_bool),
        'followers-only': r('follower_only', int_or_none),

        'r9k': r('r9k_mode', _parse_bool),
        'slow': r('slow_mode', int_or_none),
        'subs-only': r('subscriber_only', _parse_bool),
        'rituals': r('rituals_enabled', _parse_bool),

        # USERNOTICE
        'system-msg': r('system_message', _decode_pseudo_BNF),

        # (Commands)
        # HOSTTARGET
        'number-of-viewers': 'number_of_viewers',

        # ban user
        'target-user-id': r('target_author_id', str_or_none),

        # USERNOTICE - other
        **_MESSAGE_PARAM_REMAPPING
    }

    _KNOWN_IRC_KEYS = {
        # banned user
        'banned_user', 'ban_type',

        # slow mode
        'seconds_to_wait',

        # follower only
        'minutes_to_follow_before_chatting',

        # parsed elsewhere
        'action_type',
        'author',
        'in_reply_to',
        'message'
    }
    _KNOWN_IRC_KEYS.update(BaseChatDownloader.get_mapped_keys(_IRC_REMAPPING))

    _ACTION_TYPE_REMAPPING = {
        # tags
        'CLEARCHAT': 'clear_chat',
        'CLEARMSG': 'delete_message',
        'GLOBALUSERSTATE': 'successful_login',
        'PRIVMSG': 'text_message',
        'ROOMSTATE': 'room_state',
        'USERNOTICE': 'user_notice',
        'USERSTATE': 'user_state',

        # commands
        'HOSTTARGET': 'host_target',
        'NOTICE': 'notice',
        'RECONNECT': 'reconnect'
    }

    # MESSAGE GROUPS FOR IRC
    _MESSAGE_GROUP_REMAPPINGS = {
        # TODO add rest of
        # https://dev.twitch.tv/docs/irc/msg-id

        'messages': {
            'highlighted-message': 'highlighted_message',
            'skip-subs-mode-message': 'send_message_in_subscriber_only_mode',
        },
        'bits': {
            'bitsbadgetier': 'bits_badge_tier',
        },
        'subscriptions': {
            'sub': 'subscription',
            'resub': 'resubscription',
            'subgift': 'subscription_gift',
            'anonsubgift': 'anonymous_subscription_gift',
            'anonsubmysterygift': 'anonymous_mystery_subscription_gift',
            'submysterygift': 'mystery_subscription_gift',
            'extendsub': 'extend_subscription',

            'standardpayforward': 'standard_pay_forward',
            'communitypayforward': 'community_pay_forward',
            'primecommunitygiftreceived': 'prime_community_gift_received',
        },
        'upgrades': {
            'primepaidupgrade': 'prime_paid_upgrade',
            'giftpaidupgrade': 'gift_paid_upgrade',
            'rewardgift': 'reward_gift',
            'anongiftpaidupgrade': 'anonymous_gift_paid_upgrade',
        },
        'raids': {
            'raid': 'raid',
            'unraid': 'unraid'
        },
        'hosts': {
            'host_on': 'start_host',
            'host_off': 'end_host',
            'bad_host_hosting': 'bad_host_hosting',
            'bad_host_rate_exceeded': 'bad_host_rate_exceeded',
            'bad_host_error': 'bad_host_error',
            'hosts_remaining': 'hosts_remaining',
            'not_hosting': 'not_hosting',

            'host_target_went_offline': 'host_target_went_offline',
        },
        'rituals': {
            'ritual': 'ritual',
        },
        'room_states': {
            # slow mode
            'slow_on': 'enable_slow_mode',
            'slow_off': 'disable_slow_mode',
            'already_slow_on': 'slow_mode_already_on',
            'already_slow_off': 'slow_mode_already_off',

            # sub only mode
            'subs_on': 'enable_subscriber_only_mode',
            'subs_off': 'disable_subscriber_only_mode',
            'already_subs_on': 'sub_mode_already_on',
            'already_subs_off': 'sub_mode_already_off',

            # emote only mode
            'emote_only_on': 'enable_emote_only_mode',
            'emote_only_off': 'disable_emote_only_mode',
            'already_emote_only_on': 'emote_only_already_on',
            'already_emote_only_off': 'emote_only_already_off',

            # r9k mode
            'r9k_on': 'enable_r9k_mode',
            'r9k_off': 'disable_r9k_mode',
            'already_r9k_on': 'r9k_mode_already_on',
            'already_r9k_off': 'r9k_mode_already_off',

            # follower only mode
            'followers_on': 'enable_follower_only_mode',
            'followers_on_zero': 'enable_follower_only_mode',  # same thing, handled in parse
            'followers_off': 'disable_follower_only_mode',
            'already_followers_on': 'follower_only_mode_already_on',
            'already_followers_on_zero': 'follower_only_mode_already_on',
            'already_followers_off': 'follower_only_mode_already_off',

        },
        'deleted_messages': {
            'msg_banned': 'banned_message',

            'bad_delete_message_error': 'bad_delete_message_error',
            'bad_delete_message_broadcaster': 'bad_delete_message_broadcaster',
            'bad_delete_message_mod': 'bad_delete_message_mod',
            'delete_message_success': 'delete_message_success',
        },
        'bans': {

            # ban
            'already_banned': 'already_banned',
            'bad_ban_self': 'bad_ban_self',
            'bad_ban_broadcaster': 'bad_ban_broadcaster',
            'bad_ban_admin': 'bad_ban_admin',
            'bad_ban_global_mod': 'bad_ban_global_mod',
            'bad_ban_staff': 'bad_ban_staff',
            'ban_success': 'ban_success',

            # unban
            'bad_unban_no_ban': 'bad_unban_no_ban',
            'unban_success': 'unban_success',

            'msg_channel_suspended': 'channel_suspended_message',

            # timeouts
            'timeout_success': 'timeout_success',

            # timeout errors
            'bad_timeout_self': 'bad_timeout_self',
            'bad_timeout_broadcaster': 'bad_timeout_broadcaster',
            'bad_timeout_mod': 'bad_timeout_mod',
            'bad_timeout_admin': 'bad_timeout_admin',
            'bad_timeout_global_mod': 'bad_timeout_global_mod',
            'bad_timeout_staff': 'bad_timeout_staff',
        },
        'mods': {
            'bad_mod_banned': 'bad_mod_banned',
            'bad_mod_mod': 'bad_mod_mod',
            'mod_success': 'mod_success',
            'bad_unmod_mod': 'bad_unmod_mod',
            'unmod_success': 'unmod_success',
            'no_mods': 'no_mods',
            'room_mods': 'room_mods',
        },
        'colours': {
            'turbo_only_color': 'turbo_only_colour',
            'color_changed': 'colour_changed',
        },
        'commercials': {
            'bad_commercial_error': 'bad_commercial_error',
            'commercial_success': 'commercial_success',
        },

        'vips': {
            'bad_vip_grantee_banned': 'bad_vip_grantee_banned',
            'bad_vip_grantee_already_vip': 'bad_vip_grantee_already_vip',
            'vip_success': 'vip_success',
            'bad_unvip_grantee_not_vip': 'bad_unvip_grantee_not_vip',
            'unvip_success': 'unvip_success',
            'no_vips': 'no_vips',
            'vips_success': 'vips_success',
        },
        'chants': {
            'crowd-chant': 'crowd_chant'
        },
        'charity': {
            'charity': 'charity'
        },
        'other': {
            'cmds_available': 'cmds_available',
            'unrecognized_cmd': 'unrecognized_cmd',
            'no_permission': 'no_permission',
            'msg_ratelimit': 'rate_limit_reached_message',
        }
    }

    # MESSAGE GROUPS FOR VOD/CLIPS
    _MESSAGE_GROUPS = {
        'messages': [
            'text_message'
        ],
        'bans': [
            'ban_user'
        ],
        'deleted_messages': [
            'delete_message'
        ],
        'hosts': [
            'host_target'
        ],
        'room_states': [
            'room_state'
        ],
        'user_states': [
            'user_state'
        ],
        'notices': [
            'user_notice',
            'notice',
            'successful_login'
        ],
        'chants': [
            'crowd_chant'
        ],
        'other': [
            'clear_chat',
            'reconnect'
        ]
    }

    _MESSAGE_TYPE_REMAPPING = {}
    for _message_group in _MESSAGE_GROUP_REMAPPINGS:
        _value = _MESSAGE_GROUP_REMAPPINGS[_message_group]
        _MESSAGE_TYPE_REMAPPING.update(_value)

        if _message_group not in _MESSAGE_GROUPS:
            _MESSAGE_GROUPS[_message_group] = []
        _MESSAGE_GROUPS[_message_group] += list(_value.values())

    _SUBSCRIBER_BADGE_INFO = {}  # local cache for subscriber badge info
    _SUBSCRIBER_BADGE_URL = 'https://badges.twitch.tv/v1/badges/channels/{}/display'

    def _update_subscriber_badge_info(self, channel_id):
        # print('updated sub badges')
        url = self._SUBSCRIBER_BADGE_URL.format(channel_id)

        # only get if not in dict
        channel_id = int(channel_id)  # ensure integer
        if channel_id not in self._SUBSCRIBER_BADGE_INFO:
            self._SUBSCRIBER_BADGE_INFO[channel_id] = self._session_get_json(
                url).get('badge_sets') or {}

    @ staticmethod
    def _parse_item(item, offset):
        info = {}

        for key in item:
            r.remap(info, TwitchChatDownloader._COMMENT_REMAPPING,
                    key, item[key])

        if 'time_in_seconds' in info:
            info['time_in_seconds'] -= offset
            info['time_text'] = seconds_to_time(int(info['time_in_seconds']))

        channel_id = item.get('channel_id')

        # author_badges

        badges = info.pop('author_badges', None)
        if badges:
            info['author']['badges'] = list(map(lambda x: TwitchChatDownloader._parse_badge_info(
                x.get('_id'), x.get('version'), channel_id), badges))

        user_notice_params = info.pop('user_notice_params', {})

        for key in user_notice_params:
            r.remap(info, TwitchChatDownloader._MESSAGE_PARAM_REMAPPING,
                    key, user_notice_params[key], True)

        # TODO add user colour to author dict
        # TODO check this works
        # author_colour

        BaseChatDownloader._move_to_dict(info, 'author')

        original_message_type = info.get('message_type')
        if original_message_type:
            TwitchChatDownloader._set_message_type(
                info, original_message_type)
        else:
            info['message_type'] = 'text_message'

        # remove profile_image_url if present
        info.pop('profile_image_url', None)

        return info

    _OPERATION_HASHES = {
        'StreamMetadata': '1c719a40e481453e5c48d9bb585d971b8b372f8ebb105b17076722264dfa5b3e',
        'BrowsePage_Popular': 'c3322a9df3121f437182beb5a75c2a8db9a1e27fa57701ffcae70e681f502557',
        'ChannelVideoShelvesQuery': 'fb663273aa958ebe2f58d5fcb3aacc112d67ebfd7f414b095c5d1498d21aad92',
        'ClipsCards__User': 'b73ad2bfaecfd30a9e6c28fada15bd97032c83ec77a0440766a56fe0bd632777',
        'VideoMetadata': '226edb3e692509f727fd56821f5653c05740242c82b0388883e0c0e75dcbf687',
        'FilterableVideoTower_Videos': 'a937f1d22e269e39a03b509f65a7490f9fc247d7f83d6ac1421523e3b68042cb',

        # Not used yet
        # 'CollectionSideBar': '27111f1b382effad0b6def325caef1909c733fe6a4fbabf54f8d491ef2cf2f14',
        # 'ChannelCollectionsContent': '07e3691a1bad77a36aba590c351180439a40baefc1c275356f40fc7082419a84',
        # 'ComscoreStreamingQuery': 'e1edae8122517d013405f237ffcc124515dc6ded82480a88daef69c83b53ac01',
        # 'VideoPreviewOverlay': '3006e77e51b128d838fa4e835723ca4dc9a05c5efd4466c1085215c6e437e65c',
    }

    def _download_base_gql(self, ops):
        return self._session_post(self._GQL_API_URL, json=ops, headers={
            'Content-Type': 'text/plain;charset=UTF-8',
            'Client-ID': self._CLIENT_ID
        }).json()

    def _download_gql(self, ops):
        for op in ops:
            op['extensions'] = {
                'persistedQuery': {
                    'version': 1,
                    'sha256Hash': self._OPERATION_HASHES[op['operationName']],
                }
            }
        return self._download_base_gql(ops)

    _USER_REMAPPING = {
        'id': 'id',
        'login': 'name',
        'displayName': 'display_name',
        'profileImageURL': 'profile_image_url',
        'primaryColorHex': 'colour'
    }

    @staticmethod
    def _parse_user(item):
        if isinstance(item, dict):
            return r.remap_dict(item, TwitchChatDownloader._USER_REMAPPING)
        return None

    _GAME_REMAPPING = {
        'id': 'id',
        'name': 'name',
        'displayName': 'display_name',
        'boxArtURL': 'box_art_url'
    }

    @staticmethod
    def _parse_game(item):
        if isinstance(item, dict):
            return r.remap_dict(item, TwitchChatDownloader._GAME_REMAPPING)
        return None

    _CLIP_REMAPPING = {
        'id': r('id', str_or_none),
        'slug': 'slug',
        'url': 'url',
        'embedURL': 'embed_url',
        'title': 'title',
        'viewCount': 'views',
        'language': 'language',
        'curator': r('curator', _parse_user),
        'game': r('game', _parse_game),
        'language': 'language',
        'broadcaster': r('broadcaster', _parse_user),

        'thumbnailURL': 'thumbnail_url',
        'createdAt': r('created_at', timestamp_to_microseconds),
        'durationSeconds': 'duration',


    }

    def get_user_clips(self, username, limit=100, filter_by='LAST_WEEK'):
        # filter_by:
        # LAST_WEEK
        # LAST_DAY
        # LAST_MONTH
        # ALL_TIME

        # offset= 20# , offset = 0

        remaining_count = limit
        offset = 0
        while True:
            num_to_get = max(min(remaining_count, 100), 0)  # in this call
            if num_to_get <= 0:
                break

            query = [{
                'operationName': 'ClipsCards__User',
                'variables': {
                    'cursor': base64.b64encode(str(offset).encode()).decode(),
                    'login': username,
                    'limit': num_to_get,
                    'criteria': {
                        'filter': filter_by
                    }
                }
            }]
            info = self._download_gql(query)
            if not info:
                break

            clips = info[0]['data']['user']['clips']

            edges = clips['edges']
            remaining_count -= len(edges)

            for edge in edges:
                node = edge['node'] or {}
                yield r.remap_dict(node, TwitchChatDownloader._CLIP_REMAPPING)

            if not clips['pageInfo']['hasNextPage']:
                break

    _VIDEO_REMAPPING = {
        'id': r('id', str_or_none),
        'animatedPreviewURL': 'animated_preview_url',
        'game': r('game', _parse_game),

        'lengthSeconds': 'duration',

        'owner': r('owner', _parse_user),

        'previewThumbnailURL': 'preview_thumbnail_url',

        'publishedAt': r('published_at', timestamp_to_microseconds),

        'title': 'title',
        'viewCount': 'views',
        'resourceRestriction': 'resource_restriction'

        # 'contentTags': 'tags'
    }

    def get_user_videos(self, username, limit=None, video_type=None, sort='TIME'):
        # Name -> broadcastType

        # Collections (ChannelCollectionsContent)
        #

        # Past broadcasts -> ARCHIVE
        # Highlights -> HIGHLIGHT
        # Uploads -> UPLOAD
        # Past premieres -> PAST_PREMIERE
        # All Videos -> null

        # Sort -> videoSort
        # VIEWS
        # TIME
        if limit is None:
            limit = float('inf')

        remaining_count = limit
        cursor = None

        while True:
            num_to_get = max(min(remaining_count, 30), 0)  # in this call
            if num_to_get <= 0:
                break

            query = [{
                'operationName': 'FilterableVideoTower_Videos',
                'variables': {
                    'limit': num_to_get,
                    'channelOwnerLogin': username,
                    'broadcastType': video_type,
                    'videoSort': sort
                }
            }]
            if cursor:
                query[0]['variables']['cursor'] = cursor

            info = self._download_gql(query)
            if not info:
                break

            videos = multi_get(info, 0, 'data', 'user', 'videos')

            if not videos:
                break

            edges = videos['edges']
            remaining_count -= len(edges)

            for edge in edges:
                cursor = edge.get('cursor')
                node = edge.get('node')
                if not node:
                    continue
                yield r.remap_dict(node, TwitchChatDownloader._VIDEO_REMAPPING)

            if not videos['pageInfo']['hasNextPage']:
                break

    def get_featured_videos(self, username):
        query = [{
            'operationName': 'ChannelVideoShelvesQuery',
            'variables': {
                'channelLogin': username,
                'first': 5
            }
        }]
        edges = self._download_gql(
            query)[0]['data']['user']['videoShelves']['edges']
        # print(len(edges))
        return edges

    _LIVESTREAM_REMAPPING = {

        'id': r('id', str_or_none),
        'title': 'title',
        'viewersCount': 'viewers',

        'previewImageURL': 'preview_image_url',
        'broadcaster': r('broadcaster', _parse_user),
        'game': r('game', _parse_game),


        # 'tags': 'tags'

        'type': 'type'  # e.g. 'live'
    }

    def get_top_livestreams(self, limit=30):
        remaining_count = limit

        cursor = ''

        while True:
            num_to_get = max(min(remaining_count, 30), 0)
            if num_to_get <= 0:
                break
            # print('num_to_get', num_to_get)

            query = [{
                'operationName': 'BrowsePage_Popular',
                'variables': {
                    'limit': num_to_get,
                    'cursor': cursor,
                    'platformType': 'all',
                    'options': {
                        'sort': 'VIEWER_COUNT'  # RECENT, RELEVANCE, VIEWER_COUNT_ASC
                    },
                    'sortTypeIsRecency': False
                }
            }]
            edges = self._download_gql(query)[0]['data']['streams']['edges']

            cursor = edges[-1]['cursor']
            remaining_count -= num_to_get

            for edge in edges:
                node = edge['node'] or {}
                yield r.remap_dict(node, TwitchChatDownloader._LIVESTREAM_REMAPPING)

    _TWITCH_HOME = 'https://www.twitch.tv'
    _TWITCH_VIDEOS = 'https://www.twitch.tv/videos'

    def generate_urls(self, livestream_limit, vod_limit, clip_limit, **kwargs):
        # max_tests = livestream_limit + livestream_limit*(vod_limit+clip_limit)

        num_vods = math.ceil(
            vod_limit/livestream_limit) if livestream_limit > 0 else vod_limit
        num_clips = math.ceil(
            clip_limit/livestream_limit) if livestream_limit > 0 else clip_limit

        livestreams = self.get_top_livestreams(livestream_limit)
        for livestream in livestreams:
            name = livestream['broadcaster']['name']

            # e.g. https://www.twitch.tv/shroud
            yield f'{self._TWITCH_HOME}/{name}'

            vods = self.get_user_videos(name, num_vods)
            for vod in vods:
                # e.g. https://www.twitch.tv/videos/12345678
                yield f"{self._TWITCH_VIDEOS}/{vod['id']}"

            clips = self.get_user_clips(name, num_clips)
            for clip in clips:
                # e.g. https://clips.twitch.tv/FastThankfulLobsterEagleEye-SFi4SJWaTkAYu-B3
                yield clip['url']

    # offset and max_duration are used by clips

    def _get_chat_messages_by_vod_id(self, vod_id, params, max_duration, offset=None):

        # twitch does not provide messages before the stream starts,
        # so we default to a start time of 0
        start_time = ensure_seconds(
            params.get('start_time'), 0)

        e_time = params.get('end_time')
        if offset is None:  # is a vod
            offset = 0
            end_time = ensure_seconds(e_time)
            content_offset_seconds = min(start_time, max_duration)

        else:  # is a clip
            # do not allow for end_time to be None
            end_time = ensure_seconds(e_time, max_duration)
            content_offset_seconds = (start_time or 0) + offset

        max_attempts = params.get('max_attempts')

        messages_groups_to_add = params.get('message_groups') or []
        messages_types_to_add = params.get('message_types') or []

        api_url = self._API_TEMPLATE.format(vod_id, self._CLIENT_ID)

        message_count = 0
        # do not need inactivity timeout (not live)
        cursor = ''
        while True:
            url = f'{api_url}&cursor={cursor}&content_offset_seconds={content_offset_seconds}'

            for attempt_number in attempts(max_attempts):
                try:
                    info = self._session_get_json(url)
                    break
                except (JSONDecodeError, RequestException) as e:
                    self.retry(attempt_number, error=e, **params)

            error_message = multi_get(info, 'error', 'message')

            if error_message:
                raise TwitchError(error_message)

            comments = info.get('comments') or []
            for comment in comments:
                data = self._parse_item(comment, offset)

                # test for missing keys
                missing_keys = data.keys() - TwitchChatDownloader._KNOWN_COMMENT_KEYS

                if missing_keys:
                    debug_log(
                        f'Missing keys found: {missing_keys}',
                        f'Original data: {comment}',
                        f'Parsed data: {data}',
                        comment.keys(),
                        TwitchChatDownloader._KNOWN_COMMENT_KEYS
                    )

                time_in_seconds = data.get('time_in_seconds', 0)

                before_start = start_time is not None and time_in_seconds < start_time
                after_end = end_time is not None and time_in_seconds > end_time

                if before_start:  # still getting to messages
                    continue
                elif after_end:  # after end
                    return  # while actually searching, if time is invalid

                to_add = self._must_add_item(
                    data,
                    self._MESSAGE_GROUPS,
                    messages_groups_to_add,
                    messages_types_to_add
                )

                if not to_add:
                    continue

                message_count += 1
                yield data

            log('debug', f'Total number of messages: {message_count}')

            cursor = info.get('_next')

            if not cursor:
                return

    def _get_chat_by_vod_id(self, match, params):
        return self.get_chat_by_vod_id(match.group('id'), params)

    def get_chat_by_vod_id(self, vod_id, params):
        max_attempts = params.get('max_attempts')

        query = [{
            'operationName': 'VideoMetadata',
            'variables': {
                'channelLogin': '',
                'videoID': vod_id
            }
        }]

        for attempt_number in attempts(max_attempts):
            try:
                video = self._download_gql(query)[0]['data']['video']
                break
            except (JSONDecodeError, RequestException) as e:
                self.retry(attempt_number, error=e, **params)

        if not video:
            raise VideoUnavailable(
                "Sorry. Unless you've got a time machine, that content is unavailable.")
        title = video.get('title')
        duration = video.get('lengthSeconds')

        # print('duration', duration)

        channel_id = multi_get(video, 'owner', 'id')
        self._update_subscriber_badge_info(channel_id)

        return Chat(
            self._get_chat_messages_by_vod_id(
                vod_id, params, duration),
            title=title,
            duration=duration,
            status='past',
            video_type='video',
            id=vod_id
        )

    def _get_chat_by_clip_id(self, match, params):
        return self.get_chat_by_clip_id(match.group('id'), params)

    def get_chat_by_clip_id(self, clip_id, params):

        max_attempts = params.get('max_attempts')

        query = {
            'query': '{ clip(slug: "%s") { broadcaster { id } video { id createdAt } createdAt durationSeconds videoOffsetSeconds title url slug } }' % clip_id,
        }

        for attempt_number in attempts(max_attempts):
            try:
                clip = self._download_base_gql(query)['data']['clip']
                break
            except (JSONDecodeError, RequestException) as e:
                self.retry(attempt_number, error=e, **params)

        vod_id = multi_get(clip, 'video', 'id')

        if vod_id is None:
            raise NoChatReplay(
                "This clip's past broadcast has expired and chat replay is no longer available.")

        offset = clip.get('videoOffsetSeconds')

        duration = clip.get('durationSeconds')
        title = f"{clip.get('title')} ({clip_id})"

        channel_id = multi_get(clip, 'broadcaster', 'id')
        self._update_subscriber_badge_info(channel_id)

        return Chat(
            self._get_chat_messages_by_vod_id(
                vod_id, params, duration, offset),
            title=title,
            duration=duration,
            status='past',
            video_type='clip',
            id=clip_id
        )

    _MESSAGE_REGEX = re.compile(
        r'^@(.+?(?=\s+:)).*tmi\.twitch\.tv\s+(\S+)(?:.+#\S+)?(?:.:)*([^\r\n]*)', re.MULTILINE)
    # Groups:
    # 1. Tag info
    # 2. Action type
    # 3. Message

    # A full list can be found here: https://badges.twitch.tv/v1/badges/global/display

    _BADGE_KEYS = ('title', 'description', 'image_url_1x',
                   'image_url_2x', 'image_url_4x', 'click_action', 'click_url')
    _BADGE_ID_REGEX = r'v1/([^/]+)/'

    @staticmethod
    def _parse_badge_info(name, version, channel_id):
        new_badge = {
            'name': replace_with_underscores(name),
            'version': int_or_none(version, version)
        }

        # prioritise custom emotes (e.g. subscriber and bits)
        channel_id = int(channel_id)
        new_badge_info = multi_get(
            TwitchChatDownloader._SUBSCRIBER_BADGE_INFO, channel_id, name, 'versions', version) or multi_get(
            TwitchChatDownloader._BADGE_INFO, name, 'versions', version)

        if new_badge_info:
            for key in TwitchChatDownloader._BADGE_KEYS:
                new_badge[key] = new_badge_info.get(key)

            image_urls = [
                (new_badge.pop(f'image_url_{i}x', ''), i * 18) for i in (1, 2, 4)]
            if image_urls:
                new_badge['icons'] = []

            for image_url, size in image_urls:
                new_badge['icons'].append(Image(image_url, size, size).json())

            if image_urls:
                badge_id = re.search(
                    TwitchChatDownloader._BADGE_ID_REGEX, image_urls[0][0] or '')
                if badge_id:
                    new_badge['id'] = badge_id.group(1)

        return new_badge

    @staticmethod
    def _parse_irc_badges(badges, channel_id):
        info = []
        if not badges:
            return info

        for badge in badges.split(','):
            split = badge.split('/', 1)
            key_length = len(split)
            if key_length == 1:
                # If there's no /, we assign a value of None (null).
                split.append(None)

            info.append(TwitchChatDownloader._parse_badge_info(
                split[0], split[1], channel_id))
        return info

    @staticmethod
    def _set_message_type(info, original_message_type):
        new_message_type = TwitchChatDownloader._MESSAGE_TYPE_REMAPPING.get(
            original_message_type)

        if new_message_type:
            info['message_type'] = new_message_type
        else:
            debug_log(
                f'Unknown message type: {original_message_type}',
                f'Parsed data: {info}'
            )

    @staticmethod
    def _add_text_for_emotes(message, emote_list):
        for emote in emote_list:
            try:
                first_location = list(
                    map(lambda x: int(x), emote['locations'][0].split('-')))
                emote['name'] = message[first_location[0]:first_location[1] + 1]
            except Exception:
                debug_log(
                    f'Invalid emote: {emote}',
                    f'Message: {message}'
                )
                continue

    @staticmethod
    def _parse_irc_item(match):
        info = {}

        split_info = match.group(1).split(';')

        for item in split_info:
            keys = item.split('=', 1)
            key_length = len(keys)
            if key_length == 1:
                # If there's no equals, we assign the tag a value of true.
                keys.append(True)
            elif key_length == 2:
                pass
            else:  # TODO never reaches this
                debug_log(
                    f'Invalid item found: {item}.',
                    f'All items: {split_info}.',
                )
                continue

            r.remap(info, TwitchChatDownloader._IRC_REMAPPING,
                    keys[0], keys[1], keep_unknown_keys=True, replace_char_with_underscores='-')

        message_match = match.group(3)
        if message_match:
            info['message'] = remove_prefixes(message_match, '\u0001ACTION ')

            emotes = info.pop('emotes', None)
            if emotes:
                TwitchChatDownloader._add_text_for_emotes(
                    info['message'], emotes)
                info['emotes'] = emotes

        author_badge_metadata = info.pop('author_badge_metadata', [])
        author_badges = info.pop('author_badges', [])

        info['author_badges'] = TwitchChatDownloader._parse_irc_badges(
            author_badges, info.get('channel_id'))

        badge_metadata = TwitchChatDownloader._parse_irc_badges(
            author_badge_metadata, info.get('channel_id'))

        subscriber_badge = next(
            (x for x in info['author_badges'] if x.get('name') == 'subscriber'), None)
        subscriber_badge_metadata = next(
            (x for x in badge_metadata if x.get('name') == 'subscriber'), None)
        if subscriber_badge and subscriber_badge_metadata:
            subscriber_badge['months'] = int_or_none(
                subscriber_badge_metadata['version'], 0)

        author_display_name = info.get('author_display_name')
        if author_display_name:
            info['author_name'] = author_display_name.lower()

        in_reply_to = BaseChatDownloader._move_to_dict(info, 'in_reply_to')

        BaseChatDownloader._move_to_dict(in_reply_to, 'author')
        BaseChatDownloader._move_to_dict(info, 'author')

        original_action_type = match.group(2)

        if original_action_type:
            new_action_type = TwitchChatDownloader._ACTION_TYPE_REMAPPING.get(
                original_action_type)
            if new_action_type:
                info['action_type'] = new_action_type
            else:
                # unknown action type
                info['action_type'] = original_action_type
                debug_log([
                    f"Unknown action type: {info['action_type']}",
                    match,
                    info
                ])

        original_message_type = info.get('message_type')
        if original_message_type:
            TwitchChatDownloader._set_message_type(
                info, original_message_type)
        else:
            info['message_type'] = info['action_type']

        if original_action_type == 'CLEARCHAT':
            if message_match:  # is a ban
                info['message_type'] = 'ban_user'
                info['ban_type'] = 'timeout' if info.get(
                    'ban_duration') else 'permanent'
                info['banned_user'] = info.pop('message', '')

            else:  # did /clearchat
                pass

        follower_only = info.get('follower_only')
        if follower_only:
            info['follower_only'] = follower_only != -1
            if follower_only > 0:
                info['minutes_to_follow_before_chatting'] = follower_only

        slow_mode = info.get('slow_mode')
        if slow_mode is not None:
            if slow_mode != 0:
                info['slow_mode'] = True
                info['seconds_to_wait'] = slow_mode
            else:
                info['slow_mode'] = False

        # TODO
        # :tmi.twitch.tv HOSTTARGET #gothamchess :anna_chess 6612
        return info

    def _get_chat_messages_by_stream_id(self, stream_id, params):
        max_attempts = params.get('max_attempts')

        message_receive_timeout = params.get('message_receive_timeout')

        buffer_size = params.get('buffer_size')

        messages_groups_to_add = params.get('message_groups') or []
        messages_types_to_add = params.get('message_types') or []

        def create_connection():
            for attempt_number in attempts(max_attempts):
                try:
                    irc = TwitchChatIRC()
                    irc.set_timeout(message_receive_timeout)
                    irc.join_channel(stream_id)
                    return irc
                except (socket.gaierror, ConnectionRefusedError) as e:
                    self.retry(attempt_number, error=e, **params)

        twitch_chat_irc = create_connection()

        last_ping_time = time.time()

        # TODO make this a param
        ping_every = 60  # how often to ping the server

        readbuffer = ''

        message_count = 0

        try:
            while True:

                try:
                    new_info = twitch_chat_irc.recv(buffer_size)

                    if not new_info:
                        raise ConnectionError('Lost connection, reconnecting.')

                    readbuffer += new_info

                    if self._PING_TEXT in readbuffer:
                        twitch_chat_irc.send_raw(self._PONG_TEXT)

                    matches = list(self._MESSAGE_REGEX.finditer(readbuffer))
                    full_readbuffer = readbuffer.endswith('\r\n')
                    if matches:
                        if not full_readbuffer:
                            # sometimes a buffer does not contain a full message
                            # last one is incomplete

                            span = matches[-1].span()

                            pass_on = readbuffer[span[0]:]

                            # check whether message was cut off
                            if '\r\n' in pass_on:  # last message not matched
                                # only pass on incomplete message

                                # readbuffer[span[1]:]
                                pass_on = pass_on[span[1] - span[0]:]

                            # actual message cut off (matched, but not complete)
                            else:
                                # remove the last match from being processed (as it is incomplete)
                                matches.pop()

                            # pass remaining information to next attempt
                            readbuffer = pass_on

                        else:
                            # the whole readbuffer was read correctly.
                            # reset the readbuffer
                            readbuffer = ''

                        for match in matches:

                            data = self._parse_irc_item(match)

                            # test for missing keys
                            missing_keys = data.keys() - TwitchChatDownloader._KNOWN_IRC_KEYS

                            if missing_keys:
                                debug_log(
                                    f'Missing keys found: {missing_keys}',
                                    f'Original data: {match.groups()}',
                                    f'Parsed data: {data}'
                                )
                            # check whether to skip this message or not, based on its type

                            to_add = self._must_add_item(
                                data,
                                self._MESSAGE_GROUPS,
                                messages_groups_to_add,
                                messages_types_to_add
                            )

                            if not to_add:
                                continue

                            message_count += 1
                            yield data

                        log('debug',
                            f'Total number of messages: {message_count}')

                    elif full_readbuffer:
                        # No matches, but data has been read successfully.
                        # This means that we can safely reset the readbuffer.
                        # This is used to periodically reset the readbuffer,
                        # to avoid a massive buffer from forming.

                        # never pause
                        log('debug',
                            f'No matches found in "\n{readbuffer.strip()}\n"')
                        readbuffer = ''

                    current_time = time.time()

                    time_since_last_ping = current_time - last_ping_time

                    if time_since_last_ping > ping_every:
                        twitch_chat_irc.send_raw('PING')
                        last_ping_time = current_time

                except socket.timeout:
                    pass  # Allows for keyboard interrupts

                except ConnectionError:
                    # Close old connection
                    twitch_chat_irc.close_connection()

                    # Create a new connection
                    twitch_chat_irc = create_connection()

        finally:
            twitch_chat_irc.close_connection()

    def _get_chat_by_stream_id(self, match, params):
        return self.get_chat_by_stream_id(match.group('id'), params)

    def get_chat_by_stream_id(self, stream_id, params):

        max_attempts = params.get('max_attempts')

        query = [{
            'operationName': 'StreamMetadata',
            'variables': {'channelLogin': stream_id.lower()}
        }]

        for attempt_number in attempts(max_attempts):
            try:
                stream_info = self._download_gql(query)[0]['data']['user']
                break
            except (JSONDecodeError, RequestException) as e:
                self.retry(attempt_number, error=e, **params)

        if not stream_info:
            raise UserNotFound(f'Unable to find user: "{stream_id}"')

        is_live = multi_get(stream_info, 'stream', 'type') == 'live'
        channel_id = multi_get(stream_info, 'channel', 'id')
        title = multi_get(stream_info, 'lastBroadcast',
                          'title') if is_live else stream_id

        self._update_subscriber_badge_info(channel_id)

        return Chat(
            self._get_chat_messages_by_stream_id(
                stream_id, params),
            title=title,
            duration=None,
            status='live' if is_live else 'upcoming',  # Always live or upcoming
            video_type='video',
            id=stream_id
        )

    # # e.g. 'https://www.twitch.tv/spamfish/videos?filter=all'
    # _VALID_VIDEOS_URL = r'https?://(?:(?:www|go|m)\.)?twitch\.tv/(?P<id>[^/]+)/(?:videos|profile)'

    # # e.g. 'https://www.twitch.tv/vanillatv/clips?filter=clips&range=all'
    # _VALID_VIDEO_CLIPS_URL = r'https?://(?:(?:www|go|m)\.)?twitch\.tv/(?P<id>[^/]+)/(?:clips|videos/*?\?.*?\bfilter=clips)'

    # # e.g. 'https://www.twitch.tv/collections/wlDCoH0zEBZZbQ'
    # _VALID_COLLECTIONS_URL = r'https?://(?:(?:www|go|m)\.)?twitch\.tv/collections/(?P<id>[^/]+)'



"""Lists the sites that are supported"""

from .youtube import YouTubeChatDownloader
from .twitch import TwitchChatDownloader
# from .facebook import FacebookChatDownloader
from .reddit import RedditChatDownloader
from .common import BaseChatDownloader


def get_all_sites(include_parent=False):
    """Get all supported sites.

    :param include_parent: Whether to include the BaseChatDownloader, defaults to False
    :type include_parent: bool, optional
    :return: A list of all supported ChatDownloader classes
    :rtype: list
    """
    return [
        value
        for value in globals().values()
        # not the base class
        if isinstance(value, type) and issubclass(value, BaseChatDownloader) and (include_parent or value != BaseChatDownloader)
    ]



import os
import ntpath
import json
import logging


from chat_downloader import ChatDownloader
from .dataformat import * # YoutubeChatAnalytics, TwitchChatAnalytics
from .util import dprint
from urllib.parse import urlparse
from chat_downloader.sites.common import Chat

from .metadata import (
    __version__
    )

# Times in seconds dictating how granular the interval can be (how long the individual samples are)
MAX_INTERVAL = 120
MIN_INTERVAL = 1

# Can be set via CLI --debug flag
DEBUG = False

# Define the arguments for getting the chatlog using chat-downloader
chat_download_settings =  {
    "url" : None, # Set in get_chatlog_downloader()
    "message_types" : 'all',
    "output" : None, # If save-chatfile, set to that path before get_chatlog_downloader()
}

def get_chatlog_downloader(url: str):
    """
    Gets a chat-downloader generator using Xenonva's chat-downloader
    
    :param url: The URL of the past stream/VOD to download the chat from
    :param type: str
    :returns: The chatlog we have downloaded
    :rtype: chat_downloader.sites.common.Chat
    """
    # Provide url argument to the chat downloader
    chat_download_settings['url'] = url
 

    print("Getting chatlog using Xenonva's chat-downloader (https://github.com/xenova/chat-downloader)...")
    dprint(DEBUG, f"Chat download settings: {chat_download_settings}")

    try:
        chat = ChatDownloader().get_chat(
            chat_download_settings['url'], 
            message_types=[chat_download_settings['message_types']],
            output=chat_download_settings['output'])       # create a generator
    except Exception as exception:
        logging.critical("ERORR: Could not get chat: "+ str(exception))
        exit(1)

    print("Successfully retrieved chat generator:")
    print("\tTitle: %s" % (str(chat.title)))
    print("\tDuration: %s (%s seconds)" % (seconds_to_time(chat.duration), str(chat.duration)))
    print("""\033[1;31mNOTICE: Downloading chats from a url is the largest rate-limiting factor.
             \033[0;31mIf you intend to sample the data differently multiple times, consider using \033[1;33mchatfile\033[0;31m mode, or saving the chat data with \033[1;33m--save-chatfile\033[0;31m.\n\033[0m""")

    return chat

def check_chatlog_downloader_supported(chatlog: Chat, url: str):
    """
    Ensures that we are able to properly analyze the downloaded chatlog by
    enforcing the metadata matches expected values. 
    (Our own logic depends on certain things being true)

    If there is a breach of compliance, we exit (we consider this to be a fatal error)
    
    :param chatlog: The chatlog we have downloaded
    :type chatlog: chat_downloader.sites.common.Chat
    :param url: The URL of the video we have downloaded the log from
    :type url: str

    Currently we ensure:
        - The video/stream happened in the past (not currently live or scheduled)
        - The chat is downloaded from a supported platform that we have a proper
        way of parsing.
    """

    if(chatlog.status !='past'):
        logging.critical("ERROR: chat-analyzer does not support live or upcoming streams. Wait until the stream is over to use this tool!")
        exit(1)
    if(chatlog.status == None):
        logging.critical(f"ERROR: unrecognized stream status: {chatlog.status}. chat-analyzer must be able to confirm that the stream is over")
        exit(1)
    
    # NOTE: I don't currently know of a way to get the platform from the chatlog itself, so a quick operation here is a simple-enough solution
    platform = urlparse(url).netloc
    if(platform not in SUPPORTED_PLATFORMS):
        # NOTE: This is for platforms that the chat-downloader can handle but chat-analyzer can't
        logging.critical(f"ERROR: chat-analyzer does not currently support chatlogs from {platform}")
        exit(1)

def get_chatmsgs_from_chatfile(filepath: str):
    """Given a path to a chatfile (directly from Xenovas downloader or produced by 
    --save-chatfile-output) flag, extract the chat messages from the file into an array
    of chat messages.


    :param filepath: a path to a chatfile (directly from Xenovas downloader or produced by 
    --save-chatfile-output) flag
    :type filepath: str
    :returns: an array of chat messages
    :rtype: list[chat_downloader.sites.common.ChatMessage]
    """
    with open(filepath, 'r') as f:
        chat_messages = json.load(f)
    return chat_messages

def get_ChatAnalytics_from_file(filepath: str):
    """Given a path to a previous output file of this program containing analytical data,
    extract the data into a ChatAnalytics object and return it.
    
    :param filepath: a path to a previous output file of this program containing analytical data
    :type filepath: str
    :returns: a ChatAnalytics object
    :rtype: ChatAnalytics
    """

    chatAnalytics: ChatAnalytics

    with open(filepath, 'r') as f:
        jsonData = json.load(f)

        try:
            platform_from_file = jsonData['platform']
            if(platform_from_file == YOUTUBE_NETLOC):
                chatAnalytics = YoutubeChatAnalytics(-1,-1,'notseterror','notseterror','notseterror')
            elif(platform_from_file == TWITCH_NETLOC):
                chatAnalytics = TwitchChatAnalytics(-1,-1,'notseterror','notseterror','notseterror')
            else:
                logging.CRITICAL(f"Unrecognized platform when reading analytics data from file: {platform_from_file}")
                exit(1)

            for attr in dict.keys(jsonData):
                # Nested objects have to be set manually
                if(attr == 'samples'):
                    setattr(chatAnalytics, attr, [])
                    for sample_string in jsonData[attr]:
                        sample_object = Sample(-1,-1)
                        for attr2 in dict.keys(sample_string):
                            setattr(sample_object, attr2, sample_string[attr2])
                        chatAnalytics.samples.append(sample_object)
                else:
                    setattr(chatAnalytics, attr, jsonData[attr])
        except KeyError as exception:
            missing_key = str(exception)
            print(f"Could not find the {missing_key} attribute in the provided file. Please ensure that the file is a valid ChatAnalytics file previously produced by this program whose version is >= {__version__}.")
            # logging.CRITICAL(f"Error when reading analytics data from file: {str(exception)}")
            exit(1)

    return chatAnalytics

def output_json_to_file(json_obj, filepath):
    if not os.path.exists(filepath): # code block adopted from Xenova's countinous_write.py
        directory = os.path.dirname(filepath)
        if directory:  # (non-empty directory - i.e. not in current folder)
            # must make parent directory
            os.makedirs(directory, exist_ok=True)
        # open(output_filepath, 'w').close()  # create an empty file

    with open(filepath, 'w') as f:
        json.dump(json.loads(json_obj), f, ensure_ascii=False, indent=4)
    
    print(f"Successfully wrote chat analytics to {filepath}")

def run(**kwargs):
    """Runs the chat-analyzer

    :returns: The chat analytics data as a dataclass
    :rtype: dataformat.ChatAnalytics (dataformat.YoutubeChatAnalytics or dataformat.TwitchChatAnalytics)
    """

    # Interpret and extract CLI arguments from kwargs
    DEBUG = kwargs.get('debug')
    for arg in kwargs:
        value = kwargs[arg]
        dprint(DEBUG, f"analyzing arg {arg}: {value}")
    
    source = kwargs.get('source') # Is either a url or a filepath
    # Optional
    platform = kwargs.get('platform') # Required if mode==chatfile, otherwise == None
    if(platform):
        # Immediately re-parse from shorthand into official NETLOC
        platform = SUPPORTED_PLATFORMS_SHORTHANDS[platform]
    # Mode arguments
    program_mode = kwargs.get('mode') # choices=["url", "chatfile", "reanalyze"]
    save_chatfile_output = kwargs.get('save_chatfile_output')
    # Processing (Sampling) arguments
    interval = kwargs.get('interval')
    print_interval = kwargs.get('print_interval')
    # Post-processing (Analyzing) arguments
    highlight_percentile = kwargs.get('highlight_percentile')
    highlight_metric = kwargs.get('highlight_metric')
    spike_sensitivity = kwargs.get('spike_sensitivity')
    # Output
    description = kwargs.get('description')
    output_filepath = kwargs.get('output')
    # Debugging
    msg_break = kwargs.get('break')


    # Consumed by the processing/post-processing logic in dataformat.py
    process_settings = ProcessSettings(print_interval=print_interval, msg_break=msg_break, highlight_percentile=highlight_percentile, highlight_metric=highlight_metric, spike_sensitivity=spike_sensitivity)


    # Check interval argument, we check the url arg's platform in check_chatlog_supported()
    # NOTE: We double check here in addition to in CLI
    if(interval > MAX_INTERVAL or interval < MIN_INTERVAL): 
        raise ValueError(f"Sample interval must be {MIN_INTERVAL} <= interval <= {MAX_INTERVAL}")

    # Get the chat using the chat downloader and ensure that we can work with that data
    chatlog: Chat
    if(program_mode=='url'):
        if(save_chatfile_output!=None):
            chat_download_settings['output']= save_chatfile_output
            print(f"Raw chat data file will be saved to {save_chatfile_output}")
        url = source
        platform = urlparse(url).netloc
        chatlog = get_chatlog_downloader(url)
        check_chatlog_downloader_supported(chatlog, url)
    elif(program_mode=='chatfile'):
        # Have to create the Chat object manually. 
        # We just have the chat messages so we have to guess on some fields (Limitation from Xenova's chat-downloader)
        # NOTE: We could require custom chatfiles with manually imposed fields, but that destroys compatibility w/ native Xenova chat-downloader
        chat_messages = get_chatmsgs_from_chatfile(source) 
        chat_msg_iterator = iter(chat_messages)       
        chat_title = ntpath.basename(source) # We don't have the title of original vid so file name is next best thing
        chat_duration = chat_messages[-1]['time_in_seconds'] # We don't have duration so we approximate by taking the last message's timestamp
        chat_status = 'past' # (has to be, since its a chatfile)

        chatlog = Chat(chat=chat_msg_iterator, title=chat_title, duration=chat_duration, status=chat_status)
        # NOTE: platform required to provided through CLI, so don't need to set it here
        # NOTE: We assume that its a supported platform because user had to provide a platform via CLI which checks it there

    # Next section: Create the proper type of ChatAnalytics object based on the platform
    chatAnalytics: ChatAnalytics
    
    if(program_mode=='reanalyze'):
        # Create the ChatAnalytics object from the saved json file
        chatAnalytics = get_ChatAnalytics_from_file(source)
        dprint(DEBUG,f"chatanalytics object: {type(chatAnalytics)}")
        chatAnalytics.chatlog_post_process(process_settings)
    else:
        # We aren't reanalyzing a file, create the chatAnalytics object and process normally
        if(platform == YOUTUBE_NETLOC):
            chatAnalytics = YoutubeChatAnalytics(duration=chatlog.duration, interval=interval, description=description, program_version=__version__)
        elif(platform == TWITCH_NETLOC):
            chatAnalytics = TwitchChatAnalytics(duration=chatlog.duration, interval=interval, description=description, program_version=__version__)
        else:
            logging.critical(
                "ERROR: No corresponding ChatAnalytics object.\n\
                NOTE: Should have caught this in supported platforms checks earlier...")
            exit(1)
        # Now, we can process & analyze the data!
        chatAnalytics.process_chatlog(chatlog, source, process_settings)
        
    # chatAnalytics now contains all analytical data. We can print/return as ncessary
   
    json_obj = chatAnalytics.to_JSON()

    if(output_filepath==None): # If user did not specify an output filepath, use this default convention
        output_filepath =chatAnalytics.mediaTitle+'.json'
    output_json_to_file(json_obj, output_filepath)     
    
   

    return chatAnalytics




"""Set metadata for chat-analyzer"""

__title__ = 'chat-analyzer'
__program__ = 'chat_analyzer'
__summary__ = 'A tool used to process and analyze chat data from past live streams, providing summarized information about chat activity over the stream\'s lifetime.'
__author__ = 'David Fryd'
__email__ = 'info@chatanalyze.com'
__copyright__ = '2022 David Fryd'
__license__ = 'MIT'
__url__ = 'https://github.com/David-Fryd/chat-analyzer'
__website__ = 'https://chatanalyze.com'
__version__ = '1.0.0b5'




from .common import (
    BaseChatDownloader,
    Chat,
    Remapper as r
)

from ..utils.core import (
    multi_get,
    int_or_none,
    attempts,
    try_parse_json,
    try_get_first_value,
    get_title_of_webpage,
    chunks,
    ensure_seconds,
    seconds_to_time,
    regex_search
)


from ..errors import (
    SiteError,
    VideoNotFound,
    UnexpectedError,
    ChatDisabled
)

from ..debugging import (
    log,
    debug_log
)


from requests.exceptions import RequestException
from json.decoder import JSONDecodeError
import random
import json
import websocket
import time
import re


class RedditError(SiteError):
    """Raised when an error occurs with a Reddit video."""
    pass


class RedditChatDownloader(BaseChatDownloader):

    _REDDIT_HOMEPAGE = 'https://www.reddit.com'
    _INITIAL_INFO_REGEX = r'(?:window\.___r)\s*=\s*({.+?})\s*;<\/script>'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        max_attempts = 3

        for attempt_number in attempts(max_attempts):
            initial_data = self._session_get(self._REDDIT_HOMEPAGE).text

            info = regex_search(initial_data, self._INITIAL_INFO_REGEX)
            if info:
                info = try_parse_json(info)
                break
            else:
                title = get_title_of_webpage(initial_data)
                self.retry(attempt_number, text=title,
                           max_attempts=max_attempts)
                continue

        bearer = multi_get(info, 'user', 'session', 'accessToken')

        self.authed_headers = {
            'authorization': f'Bearer {bearer}',
            **self.session.headers
        }

    _NAME = 'reddit.com'

    _SITE_DEFAULT_PARAMS = {
        'format': 'default',  # TODO create reddit format
    }

    _TESTS = [
        # Different URLS
        {
            'name': 'Get chat messages from past broadcast #1',
            'params': {
                'url': 'https://www.reddit.com/rpan/r/pan/f6zbwb',
                'max_messages': 10
            },
            'expected_result': {
                'messages_condition': lambda messages: len(messages) > 0,
            }
        },


        {
            'name': 'Get chat messages from past broadcast #2',
            'params': {
                'url': 'https://www.reddit.com/rpan/f6vx11',
                'max_messages': 10
            },
            'expected_result': {
                'messages_condition': lambda messages: len(messages) > 0,
            }
        },
        {
            'name': 'Get chat messages from past broadcast #3',
            'params': {
                'url': 'https://www.reddit.com/r/pan/comments/fkzhjg',
                'max_messages': 10
            },
            'expected_result': {
                'messages_condition': lambda messages: len(messages) > 0,
            }
        },
        {
            'name': 'Get chat messages from past broadcast #4',
            'params': {
                'url': 'https://www.reddit.com/comments/fox5px',
                'max_messages': 10
            },
            'expected_result': {
                'messages_condition': lambda messages: len(messages) > 0,
            }
        },
        {
            'name': 'Get chat messages from past broadcast #5',
            'params': {
                'url': 'https://www.reddit.com/lmvsbl',
                'max_messages': 10
            },
            'expected_result': {
                'messages_condition': lambda messages: len(messages) > 0,
            }
        },
        {
            'name': 'Chat replay with start and end times #1',
            'params': {
                'url': 'https://www.reddit.com/rpan/r/pan/lmvsbl',
                'start_time': 123,
                'end_time': 456,
            },
            'expected_result': {
                'messages_condition': lambda messages: len(messages) > 0,
            }
        },
        {
            'name': 'Chat replay with start and end times #2',
            'params': {
                'url': 'https://www.reddit.com/rpan/r/pan/lmvsbl',
                'start_time': 600,
                'end_time': 900,
            },
            'expected_result': {
                'messages_condition': lambda messages: len(messages) > 0,
            }
        },
        # Anomalies
        {
            'name': 'Connection issues',
            'params': {
                'url': 'https://www.reddit.com/r/talentShow/comments/ofkm5s/carving_a_wood_vase/',
                'timeout': 5
            }
        },
        {
            'name': "This broadcast programme is off-air (has not started or has ended)",
            'params': {
                'url': 'https://www.reddit.com/rpan/r/RedditMasterClasses/olw6ww',
                'timeout': 5
            }
        },

        # Subreddits:
        {
            'name': 'Get chat of top livestream/post for subreddit #1',
            'params': {
                'url': 'https://www.reddit.com/rpan/r/pan',
                'timeout': 5
            }
        },
        {
            'name': 'Get chat of top livestream/post for subreddit #2',
            'params': {
                'url': 'https://www.reddit.com/r/RedditSessions',
                'timeout': 5
            }
        }
    ]

    # Regex provided by youtube-dl

    _VALID_URLS = {
        '_get_chat_by_subreddit_id': r'(?:https?://(?:[^/]+\.))reddit\.com/(?:(?:rpan/r|r)/)+(?P<id>[^/?#&\s]+)/?$',
        '_get_chat_by_post_id': r'(?:https?://(?:[^/]+\.))reddit\.com/(?:(?:rpan/r/[^/]+|rpan|r/[^/]+/comments|rpan|comments)/)*(?P<id>[^/?#&\s]+)',
    }

    _REMAPPING = {

        # Shared
        'author': 'author_name',
        'name': 'message_id',

        'link_id': 'post_id',

        'author_fullname': 'author_id',
        'subreddit_id': 'subreddit_id',
        'subreddit_name_prefixed': 'subreddit_name_prefixed',
        'subreddit': 'subreddit',

        'body': 'message',
        'body_html': 'message_html',

        'author_flair_type': 'author_flair_type',
        'author_flair_text_color': 'author_flair_text_color',
        'author_flair_text': 'author_flair_text',
        'author_flair_background_color': 'author_flair_background_color',

        'score': 'score',
        'created_utc': r('timestamp', lambda x: int(x * 1000000)),

        # 'author_flair_template_id':null,
        # "comment_type":null,
        # "flair_css_class":"flair-None",
        # "collapsed":false,
        # "author_flair_richtext":[],
        # "associated_award":null,
        # "distinguished":null

        # Live (websockets)
        'author_icon_img': 'author_profile_img',  # TODO parse into sizes
        'author_snoovatar_img': 'author_snoovatar_img',
        'author_is_default_icon': 'author_is_default_icon',
        'author_is_nsfw_icon': 'author_is_nsfw_icon',

        'context': r('url', lambda x: RedditChatDownloader._REDDIT_HOMEPAGE+x),

        # "attribs":[],
        # "rtjson":{...},
        # "full_date":"2021-07-17T00:20:29+00:00",
        # "collapsed_in_crowd_control":false,
        # "total_comment_count":53,
        # "flair_position":"right",
        # "_id36":"h5gn4k1",
        # "author_id":65099537,

        # Past broadcasts (API)
        'permalink': r('url', lambda x: RedditChatDownloader._REDDIT_HOMEPAGE+x),
        'total_awards_received': 'total_awards_received',
        'edited': 'is_edited',

        'parent_id': 'parent_message_id',

        'ups': 'upvotes',
        'downs': 'downvotes',

        'controversiality': 'controversiality',

        # "approved_at_utc":null,
        # "mod_reason_by":null,
        # "banned_by":null,
        # "removal_reason":null,
        # "likes":null,
        # "replies":"",
        # "user_reports":[],
        # "saved":false,
        # "banned_at_utc":null,
        # "mod_reason_title":null,
        # "gilded":0,
        # "archived":false,
        # "collapsed_reason_code":null,
        # "no_follow":true,
        # "can_mod_post":false,
        # "send_replies":true,
        # "approved_by":null,
        # "mod_note":null,
        # "all_awardings":[],
        # "awarders":[],
        # "author_patreon_flair":false,
        # "created":1625701905.0,
        # "is_submitter":false,
        # "gildings":{},
        # "collapsed_reason":null,
        # "stickied":false,
        # "author_premium":false,
        # "can_gild":false,
        # "top_awarded_type":null,
        # "score_hidden":false,
        # "num_reports":null,
        # "locked":false,
        # "report_reasons":null,
        # "treatment_tags":[],
        # "collapsed_because_crowd_control":null,
        # "mod_reports":[],
        # "subreddit_type":"public",

    }

    _KNOWN_MESSAGE_TYPES = ['new_comment', 'delete_comment',
                            'remove_comment', 'update_comment_score']

    @ staticmethod
    def _parse_item(item, start_time=None):

        info = r.remap_dict(item, RedditChatDownloader._REMAPPING)

        BaseChatDownloader._move_to_dict(info, 'author')

        author_name = multi_get(info, 'author', 'name')
        if author_name:
            info['author_display_name'] = author_name

        if isinstance(start_time, (float, int)):
            info['time_in_seconds'] = (info['timestamp'] - start_time)/1e6
            info['time_text'] = seconds_to_time(info['time_in_seconds'])

        return info

    def _get_chat_by_subreddit_id(self, match, params):
        match_id = match.group('id')
        return self.get_chat_by_subreddit_id(match_id, params)

    def _get_chat_by_post_id(self, match, params):
        match_id = match.group('id')
        return self.get_chat_by_post_id(match_id, params)

    def _try_get_info(self, url, program_params, **kwargs):
        max_attempts = program_params.get('max_attempts')
        for attempt_number in attempts(max_attempts):
            try:
                return self._session_get_json(url, **kwargs)
            except (JSONDecodeError, RequestException) as e:
                self.retry(attempt_number, error=e, **program_params)

    _API_TEMPLATE = 'https://strapi.reddit.com/videos/t3_{}'

    _STATE_REMAPPING = {
        'IS_LIVE': 'live',
        'NOT_STARTED': 'upcoming',
        'ENDED': 'past'
    }

    def get_chat_by_post_id(self, post_id, params, attempt_number=0, initial_info=None):

        if initial_info is None:  # Optimisation
            initial_info = self._try_get_info(self._API_TEMPLATE.format(
                post_id), params, headers=self.authed_headers)

        status = initial_info.get('status')
        status_message = initial_info.get('status_message')
        data = initial_info.get('data')

        if status == 'success':
            chat_disabled = data.get('chat_disabled')
            if chat_disabled:
                raise ChatDisabled('Chat is disabled for this stream.')

            post_info = data.get('post')

            stream_info = data.get('stream')

            if not stream_info:
                raise RedditError(f'Stream info not found: {data}')

            title = post_info.get('title')

            start_time = (stream_info.get('hls_exists_at') or stream_info.get(
                'publish_at') or stream_info.get('update_at'))
            if start_time:
                start_time *= 1000

            socket_url = post_info.get('liveCommentsWebsocket')

            live_status = self._STATE_REMAPPING.get(stream_info.get('state'))

            chat_item = Chat(title=title,
                             status=live_status,
                             video_type='video',
                             start_time=start_time,
                             id=post_id)

            if live_status in ('live', 'upcoming') and socket_url:  # live stream

                if not socket_url.startswith('wss://') or 'wss.redditmedia.com' not in socket_url:
                    self.retry(
                        attempt_number, text=f'Invalid websocket URL: {socket_url}', **params)
                    return self.get_chat_by_post_id(post_id, params, attempt_number + 1)

                chat_item.chat = self._get_chat_messages_by_socket(
                    socket_url, params)

            else:  # replay
                chat_item.chat = self._get_chat_messages_by_post_id(
                    post_id, params, start_time)

            return chat_item

        elif status == 'failure':
            if isinstance(data, str) and 'wait' in data.lower():
                message = f'Response from Reddit: "{data}"'
                self.retry(attempt_number, text=message, **params)
                return self.get_chat_by_post_id(post_id, params, attempt_number + 1)

            raise RedditError(data)

        elif status == 'video not found':
            raise VideoNotFound('Video not found')

        else:  # Unknown
            raise UnexpectedError(f'Info: {initial_info}')

    def _get_chat_messages_by_socket(self, socket_url, params):

        message_receive_timeout = params.get('message_receive_timeout')
        max_attempts = params.get('max_attempts')

        def create_connection():
            for attempt_number in attempts(max_attempts):
                try:
                    log('debug', f'Connecting to socket: {socket_url}')
                    ws = websocket.create_connection(socket_url)

                    # timeout needed for polling (allow keyboard interrupts)
                    ws.settimeout(message_receive_timeout)
                    return ws
                except (ConnectionError, websocket.WebSocketException) as e:
                    self.retry(attempt_number, error=e, **params)

        ws = create_connection()

        try:
            while True:
                try:
                    opcode, raw_data = ws.recv_data()
                    data = json.loads(raw_data)

                    message_type = data.get('type')

                    if message_type in self._KNOWN_MESSAGE_TYPES:
                        payload = data.get('payload')

                        parsed = self._parse_item(payload)
                        parsed['message_type'] = message_type

                        yield parsed

                    else:
                        debug_log(
                            f'Unknown message type: {message_type}', data)

                except websocket.WebSocketTimeoutException:
                    pass

                except (ConnectionError, websocket.WebSocketException):
                    # Close old connection
                    ws.close()

                    # Create a new connection
                    ws = create_connection()

        finally:
            ws.close()

    _COMMENTS_API_TEMPLATE = 'https://www.reddit.com/comments/{}/.json?limit=1'  # &sort=old

    def _get_chat_messages_by_post_id(self, post_id, params, stream_start_time=None):

        # 1. Get all comment ids
        url = self._COMMENTS_API_TEMPLATE.format(post_id)
        initial_info = self._try_get_info(url, params)

        children = multi_get(initial_info, -1, 'data', 'children')

        first_id = multi_get(children, 0, 'data', 'id')
        comment_ids = multi_get(children, -1, 'data', 'children') or []

        if first_id:
            comment_ids.insert(0, first_id)

        if not comment_ids:
            return  # No comments

        comment_ids.sort()

        num_comments = len(comment_ids)
        log('debug', f'Found {num_comments} messages')

        # https://www.reddit.com/dev/api/#GET_api_info
        info_api = 'https://www.reddit.com/api/info.json?raw_json=1&id=t1_'

        chunk_size = 100

        chunk_info = list(chunks(comment_ids, chunk_size))
        num_bins = len(chunk_info)
        all_stored = [{} for x in range(num_bins)]

        def _parse_chunk(index):
            if not all_stored[index]:  # get if not stored
                url = info_api + ',t1_'.join(chunk_info[index])
                info = self._try_get_info(url, params)
                children = multi_get(info, 'data', 'children') or []
                all_stored[index] = [self._parse_item(
                    child.get('data'), stream_start_time) for child in children]

            return all_stored[index]

        start_time = ensure_seconds(
            params.get('start_time'))
        end_time = ensure_seconds(
            params.get('end_time'), float('inf'))

        start_chunk_index = 0

        if start_time is None:
            start_time = float('-inf')

        else:
            utc_start_time = start_time*1e6 + \
                (stream_start_time or time.time()*1e3)

            def _binary_search(low, high):

                if high < low:
                    return -1

                mid = (high + low) // 2
                parsed_mid_chunk = _parse_chunk(mid)

                ts_min = multi_get(parsed_mid_chunk, 0, 'timestamp') or 0
                ts_max = multi_get(parsed_mid_chunk, -1,
                                   'timestamp') or float('inf')

                if ts_min <= utc_start_time <= ts_max:
                    return mid
                elif utc_start_time < ts_min:
                    return _binary_search(low, mid - 1)
                else:  # ts_max > utc_start_time:
                    return _binary_search(mid + 1, high)

            start_chunk_index = _binary_search(0, num_bins)

        count = 0
        # Process remaining chunks
        for index, chunk in enumerate(chunk_info):
            if index < start_chunk_index:
                continue

            for item in _parse_chunk(index):
                if item['time_in_seconds'] > end_time:
                    return

                if start_time <= item['time_in_seconds']:
                    yield item
                    count += 1

            log('debug', f'Total number of messages: {count}')

    _BROADCAST_API_URL = 'https://strapi.reddit.com/broadcasts'
    _SUBREDDIT_BROADCAST_API_URL = 'https://strapi.reddit.com/r/{}/broadcasts?page_size=1'

    def get_chat_by_subreddit_id(self, subreddit_id, params, attempt_number=0):
        # Get chat of top broadcast
        initial_info = self._try_get_info(self._SUBREDDIT_BROADCAST_API_URL.format(
            subreddit_id), params, headers=self.authed_headers)

        status = initial_info.get('status')
        data = initial_info.get('data')

        if status == 'success':
            post_id = initial_info.get('next_cursor')

            data = initial_info.pop('data', [])
            if not data:
                raise RedditError('This subreddit has no broadcasts.')

            initial_info['data'] = data[0]

            return self.get_chat_by_post_id(post_id, params, initial_info=initial_info)

        elif status == 'failure':
            if isinstance(data, str) and 'wait' in data.lower():
                message = f'Response from Reddit: "{data}"'
                self.retry(attempt_number, text=message, **params)
                return self.get_chat_by_subreddit_id(subreddit_id, params, attempt_number + 1)

            raise RedditError(data)

        else:  # Unknown
            raise UnexpectedError(f'Info: {initial_info}')

    _RPAN_API_URL = 'https://www.reddit.com/r/{}/new.json'

    # RPAN subreddits
    # https://en.wikipedia.org/wiki/Reddit_Public_Access_Network
    # Broadcasting is being rolled out to more and more subreddits, but the following are the official RPAN subreddits:
    _RPAN_SUBREDDITS = ['AnimalsOnReddit', 'DistantSocializing', 'GlamourSchool',
                        'HeadlineWorthy', 'LGBT', 'ReadWithMe', 'RedditInTheKitchen',
                        'RedditMasterClasses', 'RedditSessions', 'RedditSets',
                        'ShortCircuit', 'TalentShow', 'TheArtistStudio',
                        'TheGamerLounge', 'TheYouShow', 'WhereIntheWorld', 'pan']

    # 'TheRedditStudio' - {"reason": "private", "message": "Forbidden", "error": 403}

    def generate_urls(self, **program_params):
        # TODO add sort-by-viewers option
        # TODO add options for live and past

        max_attempts = program_params.get('max_attempts')

        # Get live streams
        for attempt_number in attempts(max_attempts):
            message = None
            error = None

            try:
                info = self._session_get_json(
                    self._BROADCAST_API_URL, headers=self.authed_headers)

                data = info.get('data')

                if isinstance(data, list):  # success
                    break

                message = f'Response from Reddit: "{data}"'

            except (JSONDecodeError, RequestException) as e:
                error = e

            self.retry(attempt_number, error=error,
                       text=message, **program_params)

        for stream in data:
            yield multi_get(stream, 'post', 'url')

        # Get previous broadcasts

        limit = 100

        past_params = {
            'sort': 'old',
            'limit': limit
        }

        api_url = self._RPAN_API_URL.format(
            random.choice(self._RPAN_SUBREDDITS))
        count = 0
        while True:
            rpan_info = self._try_get_info(
                api_url, program_params, params=past_params)

            rpan_data = rpan_info.get('data')
            if not rpan_data:
                break

            children = rpan_data.get('children') or []

            for child in children:
                data = child.get('data')
                url = data.get('permalink')

                if url and 'rpan_video' in data:

                    yield self._REDDIT_HOMEPAGE + url

                    count += 1
                    if count >= limit:  # TODO to remove?
                        return

            past_params['after'] = rpan_data.get('after')
            if not past_params['after']:
                break



import os
import json
import csv
import shutil

from ..utils.core import flatten_json


class CW:
    """
    Base class for continuous file writers.
    """

    def __init__(self, file_name, overwrite=True, **kwargs):
        """Create a CW object.

        :param file_name: The name of the file to write to
        :type file_name: str
        :param overwrite: Whether to overwrite if the file already exists, defaults to True
        :type overwrite: bool, optional
        """
        self.file_name = file_name
        self.overwrite = overwrite

    def close(self):
        self.file.close()

    def write(self, item, flush=False):
        """Write a chat item to the file. This method should be implemented in subclasses.

        :param item: The chat item
        :type item: dict
        :param flush: Whether to force the file to be flushed after writing,
            defaults to False
        :type flush: bool, optional
        :raises NotImplementedError: if the method is not implemented
            and called from a subclass.
        """
        raise NotImplementedError

    def flush(self):
        self.file.flush()


class JSONCW(CW):
    """
    Class used to control the continuous writing of a list of dictionaries to a JSON file.
    """

    def __init__(self, file_name, indent=None, separator=', ', indent_character=' ', sort_keys=True, **kwargs):
        super().__init__(file_name, **kwargs)

        self.indent = indent
        self.separator = separator
        self.indent_character = indent_character
        self.sort_keys = sort_keys

        # open file for appending and reading in binary mode.
        self.file = open(self.file_name, 'rb+')

        previous_items = []  # save previous
        if not self.overwrite:  # may have other data
            try:
                previous_items = json.load(self.file)
            except json.decoder.JSONDecodeError:
                # TODO create .tmp, file shutil.copy(), self.file.read()
                pass

        self.file.truncate(0)  # empty file

        # rewrite with new formatting
        for previous_item in previous_items:
            self.write(previous_item)

    def _multiline_indent(self, text):
        padding = self.indent * \
            self.indent_character if isinstance(
                self.indent, int) else self.indent
        return ''.join(map(lambda x: padding + x, text.splitlines(True)))

    def write(self, item, flush=False):

        self.file.seek(0, os.SEEK_END)  # Go to the end of file

        to_write = json.dumps(
            item, indent=self.indent, sort_keys=self.sort_keys)
        if self.indent is not None:
            indent_padding = '\n'  # to add on a new line
            to_write = indent_padding + self._multiline_indent(to_write)
        else:
            indent_padding = ''

        if self.file.tell() == 0:  # Check if file is empty
            # If empty, write the start of an array
            self.file.write('['.encode())
        else:
            # seek to last character
            self.file.seek(-len(indent_padding) - 1, os.SEEK_END)
            self.file.write(self.separator.encode())  # Write the separator

        self.file.write(to_write.encode())  # Dump the item
        self.file.write((indent_padding + ']').encode())  # Close the array

        if flush:
            self.flush()


class CSVCW(CW):
    """
    Class used to control the continuous writing of a list of dictionaries to a CSV file.
    """

    def __init__(self, file_name, sort_keys=True, **kwargs):
        super().__init__(file_name, **kwargs)
        self.sort_keys = sort_keys
        self.file = open(self.file_name, 'a+', newline='', encoding='utf-8')

        if not self.overwrite:
            # save previous data
            self.file.seek(0)  # go to beginning of file
            csv_dict_reader = csv.DictReader(self.file)
            self.columns = list(csv_dict_reader.fieldnames or [])
            self.all_items = [dict(x) for x in csv_dict_reader]
        else:
            self.columns = []
            self.all_items = []

        self._reset_dict_writer()

    def _reset_dict_writer(self):
        self.csv_dict_writer = csv.DictWriter(
            self.file, fieldnames=self.columns)

    def write(self, item, flush=False, flatten=True):
        if flatten:
            item = flatten_json(item)
        self.all_items.append(item)

        new_columns = [column for column in item.keys()
                       if column not in self.columns]
        if new_columns:  # new column(s) found, must rewrite whole file
            self.columns += new_columns
            if self.sort_keys:
                self.columns.sort()

            self.file.truncate(0)  # empty file

            self._reset_dict_writer()  # update writer with new columns
            self.csv_dict_writer.writeheader()  # write new header
            self.csv_dict_writer.writerows(self.all_items)  # write previous
        else:
            self.csv_dict_writer.writerow(item)  # write newest item

        if flush:
            self.flush()


class JSONLCW(CW):
    """
    Class used to control the continuous writing of a JSON lines.
    """

    def __init__(self, file_name, sort_keys=True, **kwargs):
        super().__init__(file_name, **kwargs)
        self.sort_keys = sort_keys
        self.file = open(self.file_name, 'a', encoding='utf-8')

    def write(self, item, flush=False):
        print(json.dumps(item, sort_keys=self.sort_keys),
              file=self.file, flush=flush)


class TXTCW(CW):
    """
    Class used to control the continuous writing of a text to a TXT file.
    """

    def __init__(self, file_name, **kwargs):
        super().__init__(file_name, **kwargs)
        self.file = open(self.file_name, 'a', encoding='utf-8')

    def write(self, item, flush=False):
        print(item, file=self.file, flush=flush)


class ContinuousWriter:
    _SUPPORTED_WRITERS = {
        'json': JSONCW,
        'csv': CSVCW,
        'jsonl': JSONLCW,
        'txt': TXTCW
    }

    def __init__(self, file_name=None, overwrite=True, format=None, lazy_initialise=False, **kwargs):
        """Create a ContinuousWriter object.

        :param file_name: The name of the file to write to
        :type file_name: str
        :param overwrite: Whether to overwrite if the file already exists, defaults to True
        :type overwrite: bool, optional
        :param format: The output format, defaults to None (use the extension to decide)
        :type format: str, optional
        :param lazy_initialise: Skip file creation on initialisation, defaults to False.
        :type lazy_initialise: bool, optional
        """
        super().__setattr__('data', dict())
        self.file_name = file_name
        self.overwrite = overwrite
        self.format = format
        self.lazy_initialise = lazy_initialise
        self.writer = None
        self.data.update(kwargs)

        self._initialised = False
        if not self.lazy_initialise:
            self._real_init()

    def __getattr__(self, name):
        if name in self.data:
            return self.data[name]

        raise AttributeError(
            f"'ContinuousWriter' object has no attribute '{name}'")

    def __setattr__(self, key, value):
        self.data[key] = value

    def is_default(self):
        return isinstance(self.writer, TXTCW)

    def is_initialised(self):
        return self._initialised

    def _real_init(self):
        if self._initialised:
            return

        self._initialised = True

        if self.file_name is None:
            raise AttributeError('File name not set')

        if not os.path.exists(self.file_name) or self.overwrite:
            directory = os.path.dirname(self.file_name)
            if directory:  # (non-empty directory - i.e. not in current folder)
                # must make parent directory
                os.makedirs(directory, exist_ok=True)
            open(self.file_name, 'w').close()  # create an empty file

        extension = self.format or os.path.splitext(self.file_name)[
            1][1:].lower()
        writer_class = ContinuousWriter._SUPPORTED_WRITERS.get(
            extension, TXTCW)
        self.writer = writer_class(**self.data)

    def write(self, item, flush=False):
        if not self._initialised:  # create file when first item is written
            self._real_init()

        self.writer.write(item, flush)

    def __enter__(self):
        return self

    def close(self):
        self.writer.close()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



import threading
import _thread
import time
import sys


POLLING_TIME = 0.1

SP = ' '
CR = '\r'
LF = '\n'
CRLF = CR + LF


class TimeoutOccurred(Exception):
    """Thrown when a timeout has occurred"""
    pass


def echo(text):
    sys.stdout.write(text)
    sys.stdout.flush()


# Adapted from https://github.com/johejo/inputimeout

try:
    import msvcrt

    def win_timed_input(timeout, prompt, newline):
        echo(prompt)
        begin = time.monotonic()
        end = begin + timeout
        line = ''

        while time.monotonic() < end:
            if msvcrt.kbhit():
                c = msvcrt.getwche()
                if c in (CR, LF):
                    echo(CRLF)
                    return line
                if c == '\003':
                    raise KeyboardInterrupt
                if c == '\b':
                    line = line[:-1]
                    cover = SP * len(prompt + line + SP)
                    echo(CR + cover + CR + prompt + line)
                else:
                    line += c

            time.sleep(POLLING_TIME)

        if newline:
            echo(CRLF)

        raise TimeoutOccurred

    _timed_input = win_timed_input

except ImportError:
    import selectors
    import termios

    def posix_timed_input(timeout, prompt, newline):
        echo(prompt)
        sel = selectors.DefaultSelector()
        sel.register(sys.stdin, selectors.EVENT_READ)
        events = sel.select(timeout)

        if events:
            key, _ = events[0]
            return key.fileobj.readline().rstrip(LF)
        else:
            if newline:
                echo(LF)
            termios.tcflush(sys.stdin, termios.TCIFLUSH)
            raise TimeoutOccurred

    _timed_input = posix_timed_input


def timed_input(timeout=None, prompt='', newline=False, default=None):
    if timeout is None:
        return input(prompt)
    try:
        return _timed_input(timeout, prompt, newline)
    except TimeoutOccurred:
        return default


class TimedGenerator:
    """
    Add timing functionality to generator objects.

    Used to create timed-generator objects as well as add inactivity functionality
    (i.e. return if no items have been generated in a given time period)
    """

    def __init__(self, generator, timeout=None, inactivity_timeout=None, on_timeout=None, on_inactivity_timeout=None):
        self.generator = generator
        self.timeout = timeout
        self.inactivity_timeout = inactivity_timeout

        self.on_timeout = on_timeout
        self.on_inactivity_timeout = on_inactivity_timeout

        self.timer = self.inactivity_timer = None

        if self.timeout is not None:
            self.start_timer()

        if self.inactivity_timeout is not None:
            self.start_inactivity_timer()

    def start_timer(self):
        self.timer = threading.Timer(self.timeout, _thread.interrupt_main)
        self.timer.start()

    def start_inactivity_timer(self):
        self.inactivity_timer = threading.Timer(
            self.inactivity_timeout, _thread.interrupt_main)
        self.inactivity_timer.start()

    def reset_inactivity_timer(self):
        if self.inactivity_timer:
            self.inactivity_timer.cancel()
            self.start_inactivity_timer()

    def __iter__(self):
        return self

    def __next__(self):
        to_raise = None
        set_timers = [timer for timer in (
            self.timer, self.inactivity_timer) if timer is not None]

        try:
            next_item = next(self.generator)
            self.reset_inactivity_timer()
            return next_item

        except KeyboardInterrupt as e:

            if not set_timers:
                # Neither timer has been set, so we treat this
                # as a normal KeyboardInterrupt. No need to cancel
                # timers afterwards, we can exit here.
                raise e

            # get expired timers
            expired_timers = [
                timer for timer in set_timers if not timer.is_alive()]
            if expired_timers:
                # Some timer expired
                first_expired = expired_timers[0]

                to_raise = StopIteration
                function = self.on_timeout if (
                    first_expired == self.timer) else self.on_inactivity_timeout
                self._run_function(function)

            else:  # both timers are still active, user sent a keyboard interrupt
                to_raise = e

        except Exception as e:
            # Some other error. Always propogate.
            # If e is StopIteration, there are no more items to get.
            # We can close the timers before exiting
            to_raise = e

        if to_raise:  # Something happened which will cause the generator to exit, cancel timers
            for timer in set_timers:
                timer.cancel()

            raise to_raise

    def _run_function(self, function):
        if callable(function):
            function()


def interruptible_sleep(secs, poll_time=POLLING_TIME):
    start_time = time.time()

    while time.time() - start_time <= secs:
        time.sleep(poll_time)



import re
import os
import json

from ..utils.core import (
    nested_update,
    multi_get,
    microseconds_to_timestamp,
    seconds_to_time,
    time_to_seconds
)
from copy import deepcopy


from ..errors import (
    FormatNotFound,
    FormatFileNotFound
)


class ItemFormatter:
    """Class used to control the formatting of chat items."""

    _INDEX_REGEX = r'(?<!\\){(.+?)(?<!\\)}'

    # 'always_show': True (default False)

    def __init__(self, path=None):
        """Create an ItemFormatter object

        :param path: Path of the format file, defaults to None
        :type path: str, optional
        """
        default_path = os.path.join(os.path.dirname(
            os.path.realpath(__file__)), 'custom_formats.json')

        with open(default_path) as default_formats:
            self.format_file = json.load(default_formats)

        if path is not None:
            if not os.path.exists(path):
                raise FormatFileNotFound(
                    f'Format file not found: "{path}"')

            with open(path) as custom_formats:
                self.format_file.update(json.load(custom_formats))

    def _replace(self, match, item, format_object):
        """Replace a match object with

        :param match: The match object
        :type match: re.Match
        :param item: The chat item to choose the value to replace the key with
        :type item: dict
        :param format_object: The format object which defines how the
            replacement should be done
        :type format_object: dict
        :return: The replacement value as a string
        :rtype: str
        """

        split = match.group(1).split('|')

        for index in split:
            value = multi_get(item, *index.split('.'))

            if value is None:
                continue

            formatting_info = format_object.get(index)
            if formatting_info is None:
                return str(value)

            template = ''
            if isinstance(formatting_info, str):
                template = formatting_info
            elif isinstance(formatting_info, dict):
                template = formatting_info.get('template') or ''
                formatting = formatting_info.get('format')
                if formatting:
                    if index == 'timestamp':
                        value = microseconds_to_timestamp(
                            value, formatting)
                    elif index == 'time_text':
                        collapse_leading_zeroes = formatting_info.get(
                            'collapse_leading_zeroes')
                        value = seconds_to_time(time_to_seconds(
                            value), formatting, collapse_leading_zeroes)
                    else:
                        pass   # TODO add others

                # Apply separator
                separator = formatting_info.get('separator')
                if separator:
                    if index == 'author.badges':
                        value = separator.join(
                            map(lambda key: key.get('title'), value))
                    elif isinstance(value, (tuple, list)):
                        value = separator.join(
                            map(lambda x: str(x), value))
                    else:
                        pass
            else:
                pass

            return template.format(value)

        return ''  # no match, return empty

    def format(self, item, format_name='default', format_object=None):
        """Format a chat item according to a format (specified by its name),
            found in the format_object

        :param item: The chat item to be formatted
        :type item: dict
        :param format_name: The name of the format to be applied, defaults
            to 'default'
        :type format_name: str, optional
        :param format_object: The format object from which the format will
            be chosen, defaults to None
        :type format_object: dict, optional
        :return: The string representation of the chat item
        :rtype: str
        """
        default_format_object = self.format_file.get('default')
        if format_object is None:
            format_object = self.format_file.get(format_name)
            if not format_object:
                if format_name != 'default':
                    raise FormatNotFound(
                        f'Format not found: "{format_name}"')
                else:
                    format_object = default_format_object  # Set to default

        if isinstance(format_object, list):
            does_match = False

            for fmt in format_object:
                matching = fmt.get('matching')
                message_type = item.get('message_type')
                if isinstance(matching, list):
                    does_match = message_type in matching
                else:
                    does_match = matching == 'all' or message_type == matching

                if does_match:
                    format_object = fmt
                    break

            if not does_match:
                format_object = default_format_object
            # format_object = next((x for x in format_object if item.get(
            #     'message_type') in x.get('matching') or x.get('matching') == 'all'), None)

        if not format_object:
            return  # raise no format given

        inherit = format_object.get('inherit')
        if inherit:
            parent = self.format_file.get(inherit) or {}
            format_object = nested_update(deepcopy(parent), format_object)

        template = format_object.get('template') or ''
        keys = format_object.get('keys') or {}

        substitution = re.sub(self._INDEX_REGEX, lambda match: self._replace(
            match, item, keys), template)

        return substitution



"""File for defining errors"""


class ChatDownloaderError(Exception):
    """Base class for Chat Downloader errors."""


class UnexpectedError(ChatDownloaderError):
    """Raised if something unexpected happens."""

    def __init__(self, items):
        super().__init__(str(items))


class InvalidParameter(ChatDownloaderError):
    """Raised if an invalid parameter is specified."""
    pass


class RetriesExceeded(ChatDownloaderError):
    """Raised after the maximum number of retries has been reached."""
    pass


class VideoNotFound(ChatDownloaderError):
    """Raised when video cannot be found."""
    pass


class UserNotFound(ChatDownloaderError):
    """Raised when user cannot be found."""
    pass


class NoVideos(ChatDownloaderError):
    """Raised when a channel does not have any videos."""
    pass


class ParsingError(ChatDownloaderError):
    """Raised when video data cannot be parsed."""
    pass


class VideoUnavailable(ChatDownloaderError):
    """Raised when video is unavailable."""
    pass


class LoginRequired(ChatDownloaderError):
    """Raised when video is login is required (e.g. if video is private)."""
    pass


class VideoUnplayable(ChatDownloaderError):
    """Raised when video is unplayable (e.g. if video is members-only)."""
    pass


class NoChatReplay(ChatDownloaderError):
    """Raised when the video does not contain a chat replay."""
    pass


class ChatDisabled(ChatDownloaderError):
    """Raised when the chat is disabled."""
    pass


class URLNotProvided(ChatDownloaderError):
    """Raised when no url is provided."""
    pass


class InvalidURL(ChatDownloaderError):
    """Raised when the url is invalid."""
    pass


class ChatGeneratorError(ChatDownloaderError):
    """Raised when no valid generator method for a site can be found."""
    pass


class SiteError(ChatDownloaderError):
    """Raised when an error occurs with a specific site."""
    pass


class SiteNotSupported(SiteError):
    """Raised when the url is valid, but the site is not supported."""
    pass


class NoContinuation(ChatDownloaderError):
    """Raised when no continuation can be found."""
    pass


class CookieError(ChatDownloaderError):
    """Raised when an error occurs while loading a cookie file."""
    pass


class FormatError(ChatDownloaderError):
    """Raised when a formatting error occurs"""
    pass


class FormatNotFound(FormatError):
    """Raised when a specified format can not be found"""
    pass


class FormatFileNotFound(FormatError):
    """Raised when the format file can not be found"""
    pass
